#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.files_rw deals with file format conversion 
between different file standards .

it can be imported directly with:
from geodezyx import files_rw

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

#from geodezyx import utils,dt,time,np,os,re,struct,math,string,pd,copy
#from geodezyx import *
#import dateutil

#import geodezyx.megalib.geodetik as geok

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import dateutil
from io import BytesIO,StringIO
import itertools
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import os 
import pandas as pd
import shutil
import textwrap
import re

#### geodeZYX modules
from geodezyx import conv
from geodezyx import operational
from geodezyx import utils
from geodezyx.files_rw import read_logsheets

#### Import the logger
import logging
log = logging.getLogger(__name__)

##########  END IMPORT  ##########


   #_____          __  __ _____ _______                                   _
  #/ ____|   /\   |  \/  |_   _|__   __|                                 | |
 #| |  __   /  \  | \  / | | |    | |      __ _  ___ _ __   ___ _ __ __ _| |
 #| | |_ | / /\ \ | |\/| | | |    | |     / _` |/ _ \ '_ \ / _ \ '__/ _` | |
 #| |__| |/ ____ \| |  | |_| |_   | |    | (_| |  __/ | | |  __/ | | (_| | |
  #\_____/_/    \_\_|  |_|_____|  |_|     \__, |\___|_| |_|\___|_|  \__,_|_|
                                          #__/ |
                                         #|___/
#

def list_stat_in_statinfo(statinfoin):
    print("WARN : DISCONTINUED : list_stat_in_statinfo doesnt work well, should be replaced by stat_list_in_station_info !!!!")
    listtemp = []
    for l in open(statinfoin):
        if l[0] != ' ':
            continue
        if l[1:5] == '\n':
            continue
    listtemp.append(l[1:5])
    listout = utils.uniqify_list(listtemp)
    return listout


def read_station_info_solo(filein,stat,column_type="ulr"):
    """
    column_type : str
        "ulr" or "sopac"
    """
    dicout = {k: [] for k in ['Start','End','Rec','Ant',
              'AntHt','Dome','Ant N','Ant E','Vers',
              'SwVers','Rec SN','Ant SN']}

    if column_type == "sopac":
        rg_start_1 = slice(25,29)
        rg_start_2 = slice(30,33)
        rg_start_3 = slice(34,36)
        rg_start_4 = slice(37,39)
        rg_start_5 = slice(40,42)

        rg_end_1   = slice(44,48)
        rg_end_2   = slice(49,52)
        rg_end_3   = slice(53,55)
        rg_end_4   = slice(56,58)
        rg_end_5   = slice(59,61)

        rg_rec   = slice(96,114)
        rg_antht = slice(64,70)
        rg_ant   = slice(170,187)
        rg_dome  = slice(187,191)
        rg_antN  = slice(80,86)
        rg_antE  = slice(89,95)

        rg_vers     = slice(119,141)
        rg_sw_vers  = slice(141,148)
        rg_rec_sn   = slice(148,170)
        rg_ant_sn   = slice(170,187)

    elif column_type == "ulr":
        rg_start_1 = slice(25,29)
        rg_start_2 = slice(30,33)
        rg_start_3 = slice(34,36)
        rg_start_4 = slice(37,39)
        rg_start_5 = slice(40,42)

        rg_end_1   = slice(44,48)
        rg_end_2   = slice(49,52)
        rg_end_3   = slice(53,55)
        rg_end_4   = slice(56,58)
        rg_end_5   = slice(59,61)

        rg_rec   = slice(97,103)
        rg_antht = slice(64,70)
        rg_ant   = slice(134,140)
        rg_dome  = slice(142,146)
        rg_antN  = slice(80,86)
        rg_antE  = slice(89,95)

    else:
        print("ERR : check column_type")
        return None

    for l in open(filein):
        if l[1:5] == stat:
            # cas specifique du temps (repris d'une autre fct)
            f = l
            start = conv.doy2dt(int(f[rg_start_1]),
                                int(f[rg_start_2]),
                                int(f[rg_start_3]),
                                int(f[rg_start_4]),
                                int(f[rg_start_5]))
            dicout['Start'].append(start)
            if int(f[rg_end_1]) == 9999 or int(f[rg_end_2]) == 999:
                end = conv.doy2dt(2099,1,0,0,0)
            else:
                end = conv.doy2dt(int(f[rg_end_1]),
                                  int(f[rg_end_2]),
                                  int(f[rg_end_3]),
                                  int(f[rg_end_4]),
                                  int(f[rg_end_5]))
            dicout['End'].append(end)
            # Autres Donnees importantes
            dicout['Rec'].append(l[rg_rec].strip())
            dicout['AntHt'].append(float(l[rg_antht].strip()))
            dicout['Ant'].append(l[rg_ant].strip())
            dicout['Dome'].append(l[rg_dome].strip())
            dicout['Ant N'].append(l[rg_antN].strip())
            dicout['Ant E'].append(l[rg_antE].strip())

            if column_type == "sopac":
                dicout['Vers'].append(l[rg_vers].strip())
                dicout['SwVers'].append(l[rg_sw_vers].strip())
                dicout['Rec SN'].append(l[rg_rec_sn].strip())
                dicout['Ant SN'].append(l[rg_ant_sn].strip())

    return dicout

def read_station_info_solo_date(filein,stat,date,column_type="ulr"):
    DIC = read_station_info_solo(filein,stat,column_type=column_type)

    Start = DIC["Start"]
    End   = DIC["End"]

    i_good = None
    for i,(s,e) in enumerate(zip(Start,End)):
        if (s <= date) and (date <= e):
            if i_good:
                print("WARN : index of the good period has been already defined, something is weird ...")
            i_good = i

    if i_good is None:
        print('ERR : no corresponding metadata have been found in the station.info for :' )
        log.info(stat,date)
        return None

    dicout = dict()

    for k , v in DIC.items():
        dicout[k] = v[i_good]

    return dicout



#dic = read_station_info_solo('/home/pierre/Documents/stationinfo2gins/station.info.ovsg','HOUE')

def read_lfile_solo(filein,stat):
    """
    WEAK : should be improved with discontinuites management
    """
    for l in open(filein):
        X, Y, Z  = 0.,0.,0.
        vX,vY,vZ = 0.,0.,0.

        T = dt.datetime.now()

        if l[1:5] == stat:
            f = l[1:].split()

            X  = float(f[1])
            Y  = float(f[2])
            Z  = float(f[3])
            Ttmp = float(f[7])

            if np.isclose(Ttmp , 0.):
                T  = conv.convert_partial_year(2000.)
            else:
                T  = conv.convert_partial_year(Ttmp)

            if len(l) > 225: # if velocities are given (125 is arbitrary)
                vX = float(f[8])
                vY = float(f[9])
                vZ = float(f[10])
            else: # if no velocityis given
                vX = 0.
                vY = 0.
                vZ = 0.


            return X,Y,Z,T,vX,vY,vZ

    log.warning("no coords found for : %s ")
    return None



def read_pbo_vel_file_solo(velfilein,stat):
    """
    return a LIST of panda dataframe (corrsponding to different discontinuites),
    where the keywords of the pbo .vel can be used directly

    exemple :
    ftp://data-out.unavco.org/pub/products/velocity/pbo.final_nam08.vel
    """
    fil = open(velfilein)
    skiplinelis = []
    for i,l in enumerate(fil):
        if l[0] not in (' ','*'):
            skiplinelis.append(i)

    data = pd.read_table(open(velfilein),skiprows=skiplinelis,sep= ' *',header=max(skiplinelis)+1)
    #data['Ref_epoch']

    i_lis = []
    for i,s in enumerate(list(data['*Dot#'])):
        if s == stat.upper():
            i_lis.append(i)

    data2 = data.transpose()

    data_lis = []
    for i in i_lis:
        data_lis.append(data2[i])
    return data_lis


def read_globk_vel_file(velfile_in):

    p = velfile_in
    i = 0
    for l in open(p):
        i += 1
        if '(deg)' in l:
            break

    D = pd.read_table(p,skiprows=i,header=-1,delim_whitespace = True )

    D.rename(columns={0: 'Long',
                      1: 'Lat',
                      2: 'E',
                      3: 'N',
                      4: 'E_adj',
                      5: 'N_adj',
                      6: 'E_sig',
                      7: 'N_sig',
                      8: 'Rho',
                      9: 'H',
                      10: 'H_adj',
                      11: 'H_sig',
                      12: 'SITE'}, inplace=True)

    return D


def read_sinex_discontinuity_solo(snxfile,stat,PorV = 'P'):

    flagxyz = False

    soln_lis  = []
    start_lis = []
    end_lis   = []
    PorV_lis  = []
    notes_lis = []

    for line in open(snxfile):

        if re.compile('SOLUTION/DISCONTINUITY').search(line):
            flagxyz = not flagxyz
            continue

        if line[0] != ' ':
            continue

        if flagxyz == True:
            fields = line.split()

            if line[0] != ' ':
                continue

            if fields[0] == stat.upper():

                if not fields[6] == PorV:
                    continue

                soln_lis.append(int(fields[2]))
                start_lis.append(conv.datestr_sinex_2_dt(fields[4]))
                end = conv.datestr_sinex_2_dt(fields[5])
                if end == dt.datetime(1970,1,1):
                    end = dt.datetime(2099,1,1)
                end_lis.append(end)
                PorV_lis.append(fields[6])
                notes_lis.append(line.split('-')[-1].strip())

    return start_lis , end_lis , notes_lis , PorV_lis

def stat_list_in_station_info(filein):
    statlist = []
    fstatinfo = open(filein,"r+")
    for line in fstatinfo:
        if  len(line) == 0:
            log.warning('empty line in the station.info !!!')
            continue
        elif  line[0] != ' ':
            continue
        elif ' ' == line[0]:
            f = line.split()
            statlist.append(f[0])
    out = sorted(list(set(statlist)))
    return out

def convert_statinfo2eqfile(statinfoin,eqfileout):
    outfileobj = open(eqfileout,'w+')
    stat_list = stat_list_in_station_info(statinfoin)
    for stat in stat_list:
        i = 0
        startlis , endlis = read_station_info_time_solo(statinfoin,stat)
        statbisproto = stat + '_GPS'
        for s,e in zip(startlis,endlis):
            if i !=0:
                statbis_list = list(statbisproto)
                if i > 9:
                    statbis_list[5] = str(i)[0]
                    statbis_list[6] = str(i)[1]
                else:
                    statbis_list[5] = str(i)
                statbis = ''.join(statbis_list)
            else:
                statbis = statbisproto
            sy = s.year
            smo = s.month
            sd = s.day
            sh = s.hour
            smi = s.minute
            ss = s.second
            ey = e.year
            emo = e.month
            ed = e.day
            eh = e.hour
            emi = e.minute
            es = e.second
            ligne = ' rename {}     {} {:2} {:2} {:2} {:2} {:2} {:2} {:2} {:2} {:2} {:2}\n'.format(stat,statbis,sy,smo,sd,sh,smi,ey,emo,ed,eh,emi)
            outfileobj.write(ligne)
            i = i+1
    outfileobj.close()
    return None


def read_station_info_time_solo(filein,stat):
    ''' BROUILLON
        pour une station, donne les startdate et enddate d'une periode
        SORTIE : 2 x liste de dt '''

    startlis = []
    endlis = []


    for l in open(filein):

        if l[1:5] == stat:

            f = l[25:].split()

            start = conv.doy2dt(int(f[0]),int(f[1]),int(f[2]),int(f[3]),int(f[4]))
            startlis.append(start)

            if int(f[5]) == 9999 or int(f[6]) == 999:
                end = conv.doy2dt(2099,1,0,0,0)
            else:
                end = conv.doy2dt(int(f[5]),int(f[6]),int(f[7]),int(f[8]),int(f[9]))
            endlis.append(end)

    return startlis,endlis



def read_eqfile_time_solo(filein,stat):
    ''' BROUILLON
        pour une station, donne les startdate et enddate d'une periode
        SORTIE : 2 x liste de dt '''

    startlis = []
    endlis = []

    for l in open(filein):
        f = l.split()
        if f[1] == stat:

            start = dt.datetime(int(f[3]),int(f[4]),int(f[5]),int(f[6]),int(f[7]))
            startlis.append(start)

            if int(f[7]) == 9999 or int(f[8]) == 999:
                end = dt.datetime(2099,1,1,0,0)
            else:
                end = dt.datetime(int(f[8]),int(f[9]),int(f[10]),int(f[11]),int(f[12]))
            endlis.append(end)

    return startlis,endlis


def read_eqfile_as_dico(filein, delGPSfield = False):
    '''
    for an eqfile return a dico
    dic[STAT] = [disc1 , disc2 ...] where disc are
    the discont of the FIRST column

    delGPSfield manage the elimination of the line of type _GPS
    which are not regular
    EDIT : 150817 WHY ???

    '''

    startlis = []
    endlis = []

    outdico = dict()

    for l in open(filein):
        f = l.split()
        stat = f[1]
        if stat not in outdico:
            outdico[stat] = []
        if f[2][-3] == 'G' and delGPSfield:
            continue
        datelis = [f[3],f[4],f[5],f[6],f[7],f[8]]
        datelis = [int(e) for e in datelis]
        outdico[stat].append(dt.datetime(*datelis))

    return outdico

def write_eqfile_from_dico(dicoin,outdir,outfilename):
    outfilepath = os.path.join(outdir,outfilename)
    fil = open(outfilepath,'w+')

    from collections import OrderedDict
#    dicoin = OrderedDict(sorted(dicoin.items(), key=lambda dicoin: dicoin[1]))
#    dicoin = sorted(dicoin, key=dicoin.get)
    for k,v in dicoin.items():
        for i in range(len(v)):
            stat = k
            if i+2 > 9:
                statbis = stat + '_' + str(i+2) + 'S'
            else:
                statbis = stat + '_' + str(i+2) + 'PS'

            s = v[i]
            if i == len(v)-1:
                e = dt.datetime(2099,1,1)
            else:
                e = v[i+1]

            sy = s.year
            smo = s.month
            sd = s.day
            sh = s.hour
            smi = s.minute
            ss = s.second
            ey = e.year
            emo = e.month
            ed = e.day
            eh = e.hour
            emi = e.minute
            es = e.second

            ligne = ' rename {}     {} {:2} {:2} {:2} {:2} {:2} {:2} {:2} {:2} {:2} {:2}\n'.format(stat,statbis,sy,smo,sd,sh,smi,ey,emo,ed,eh,emi)
            fil.write(ligne)
    return None





#      _        _   _             _        __        ___     _____       _______ _____
#     | |      | | (_)           (_)      / _|      |__ \   / ____|   /\|__   __/ ____|
#  ___| |_ __ _| |_ _  ___  _ __  _ _ __ | |_ ___      ) | | |       /  \  | | | (___
# / __| __/ _` | __| |/ _ \| '_ \| | '_ \|  _/ _ \    / /  | |      / /\ \ | |  \___ \
# \__ \ || (_| | |_| | (_) | | | | | | | | || (_) |  / /_  | |____ / ____ \| |  ____) |
# |___/\__\__,_|\__|_|\___/|_| |_|_|_| |_|_| \___/  |____|  \_____/_/    \_\_| |_____/
#


def statname_of_catsfile(catsneu_path):
    catsneu_f = open(catsneu_path,'r')
    for line in catsneu_f:
        if 'Site' in line:
            return line.split(' ')[2]
    return 'XXXX'

def statinfo_2_cats(statinfo_path,catsneu_path):
    stat = statname_of_catsfile(catsneu_path)
    log.info('stat = ',stat)
    if stat == 'ABMF':
        return None
    stat_dic = read_station_info_solo(statinfo_path,stat)
    start = stat_dic['Start']
    antht = stat_dic['AntHt']

    outcats_path = os.path.dirname(catsneu_path) + os.path.basename(catsneu_path) + '.offset'
#    shutil.copyfile(catsneu_path,outcats_path)

    antht = [a for (s,a) in sorted(zip(start,antht))]
    start = sorted(start)
    log.info(stat_dic,stat)
    s0 = start[0]
    h0 = antht[0]
    s_stk = []
    for s, h in zip(start,antht):
        if h != h0:
            s_stk.append(s)
            h0 = h
    log.info(outcats_path)
    catsneuout_f = open(outcats_path,'w')
    catsneu_f = open(catsneu_path,'a+')
    lastline=''
    for line in catsneu_f:
        if lastline == line:
            continue
        elif 'Height' in line:
            catsneuout_f.write(line)
            for s in s_stk:
                catsneuout_f.write('# offsets : ' + str(conv.toYearFraction(s)) + ' 1\n')
        # cleaning outlier
        elif line[0] != '#':
            if (abs(float(line.split()[1])) > 1 or abs(float(line.split()[2])) > 1):
                pass
            else:
                catsneuout_f.write(line)
        else:
            catsneuout_f.write(line)
        lastline = line


    catsneuout_f.close()
    # type 1 for only vertical (cf doc)
    return None

#      _        _   _             _        __        ___          _
#     | |      | | (_)           (_)      / _|      |__ \        (_)
#  ___| |_ __ _| |_ _  ___  _ __  _ _ __ | |_ ___      ) |   __ _ _ _ __  ___
# / __| __/ _` | __| |/ _ \| '_ \| | '_ \|  _/ _ \    / /   / _` | | '_ \/ __|
# \__ \ || (_| | |_| | (_) | | | | | | | | || (_) |  / /_  | (_| | | | | \__ \
# |___/\__\__,_|\__|_|\___/|_| |_|_|_| |_|_| \___/  |____|  \__, |_|_| |_|___/
#                                                            __/ |
#                                                           |___/

#def oceload_associed(statginsfile):
#    import os
#
#    aaa = os.path.abspath(statginsfile)
#    bbb = 'outocload'
#    filearg = open('temparg','wr')
#    filearg.write(aaa + '\n')
#    filearg.write(os.path.basename(aaa) + '/' + bbb)
#
#    os.system('bash /home/psakicki/scripts/exe_loadoce < temparg')
#

def receptor_gins_corrector(inRec):
    outRec = inRec
    outRec = outRec.replace('ASHTECH','AS')
    outRec = outRec.replace('TRIMBLE','TR')
    outRec = outRec.replace('LEICA GRX1200+GNS','LEICA GRX12')
    outRec = outRec.replace('LEICA GRX1200GGPRO','LEICA GRX12')
    outRec = outRec.replace('ROGUE','RG')
    return outRec

def station_info_2_gins(statinfoin,coordfilein,outfile,
                        coordfile_type ='pbovelfile',specific_stats_lis=[],
                        ellipsoid='GRS80',station_info_columns_type="sopac"):
    """
    Convert a GAMIT station.info to a GINS stations files

    Parameters
    ----------
    statinfoin : str
        path of input station.info.
    coordfilein : str
        path of input coordinates files.
    outfile : str
         path of output GINS station file.
    coordfile_type : str, optional
        GAMIT coordinates files type : 'lfile' or 'pbovelfile'.
        The default is 'pbovelfile'.
    specific_stats_lis : list, optional
         list of specific stations. The default is [].
    ellipsoid : str, optional
        ellipsoid. The default is 'GRS80'.
    station_info_columns_type : TYPE, optional
        it exists different subtypes of station.info ...
        handles "sopac" or "ulr" yet. The default is "sopac".


    Returns
    -------
    outfile : str
         path of GINS outfile.

    """

    # Info de referentiel de la 1ère ligne
    refiel = 'i08i08'
    # Info de referentiel des lignes de réoccupations
    refchange = 'igs10'
    # code 'GRG3'
    grg = 'GRG3'
    # sigma sur les offsets
    sigoffset = 0.000
    # sigma sur les positions
    sigposi = 0.030
    # sigma sur les vitesses
    sigvit=0.005
    # Vitesses

    # ID DOMES du pays
    # ex : 971 pour la Guadeloupe
    # mettre 999 dans la plupart des cas
    # ATTENTION : string entre quotes
    #idprefix='971'
    #idprefix='990'
    #EDIT 1805, for very big station.info now idprfix is disabled
    #idprefix='990'

    # id initial, les stations auront un numero atribué de maniere décroissante
    # 97198 97197 97196 ...
    # idstatinit=99
    #EDIT 1805, for very big station.info now idstatinit is now at 99999
    idstatinit=99999

    # ======== FIN DES MODIFIABLES ========

    outEnd='000000'

    statgins_filobj=open(outfile,'w')

    if specific_stats_lis == []:
        listat = stat_list_in_station_info(statinfoin)
    else:
        listat = specific_stats_lis

    log.info(listat)

    sigvit = '{:.4f}'.format(sigvit).lstrip('0')

    header  = header_from_ellipsoid(ellipsoid)
    statgins_filobj.write(header)

    for stat in listat:

        log.info(stat)

        # On degresse l'ID quoiqu'il arrive après
        # pour conserver une homogenéité dans les pseudoDOMES
        idstatinit = idstatinit - 1

        if coordfile_type == 'lfile':
            output_lfile = None
            output_lfile = read_lfile_solo(coordfilein,stat)

            if not output_lfile:
                log.info("WARN : station_info_2_gins : skip station " , stat )
                continue
            else:
                X,Y,Z,T,vX,vY,vZ = output_lfile

        elif coordfile_type == 'pbovelfile':
            data = read_pbo_vel_file_solo(coordfilein,stat)
            if len(data) != 0:
                data = data[0]
                X = data['Ref_X']
                Y = data['Ref_Y']
                Z = data['Ref_Z']
                T = conv.MJD2dt(data['Ref_jday'])
                vX = data['dX/dt']
                vY = data['dY/dt']
                vZ = data['dZ/dt']
            else:
                log.warning('no %s found in %s',stat,coordfilein)
                continue
        else:
            raise Exception('Wrong coordfile_type')

        if X == 0:
            continue

        #EDIT 1805, for very big station.info now idprefix is disabled
        #idstat = str(idprefix) + str(idstatinit)
        idstat = str(idstatinit)

        T = T.strftime('%d%m%y')

        dicout = read_station_info_solo(statinfoin,stat,
                                        column_type=station_info_columns_type)

        n = len(dicout['Start'])
        log.info("N = %s", n)

        void='             '
        if stat == 'ABMF':
            idstat=str(97103)

        fortupsta=(idstat,'M001',stat,void,X,sigposi,Y,Z,vX,sigvit,vY,vZ,T,refiel)

        outputline="   {0} {1} {2}{3}{4:>-12.3f} ~{5:5.3f} {6:>-12.3f} ~{5:5.3f} {7:>-12.3f} ~{5:5.3f} {8:>-7.4f} ~{9} {10:>-7.4f} ~{9} {11:>-7.4f} ~{9} {12}        {13}".format(*fortupsta)

        outputline=outputline+'\n'
        statgins_filobj.write(outputline)


        for i in range(n):
            idligne  = idstat + '{:0>2d}'.format(i+1)
            outAnt   = dicout['Ant'][i]
            outAntHt = dicout['AntHt'][i]
            Xant,Yant,Zant = conv.ENU2XYZ_legacy(0,0,outAntHt,X,Y,Z)
            lastEnd  = outEnd
            outStart = dicout['Start'][i].strftime('%d%m%y')
            outEnd   = dicout['End'][i].strftime('%d%m%y')

            if outStart == lastEnd:
                outStart = (dicout['Start'][i] + dt.timedelta(days=1)).strftime('%d%m%y')

            if dicout['End'][i] == dt.datetime(2099,1,1,0,0,0):
                outEnd = '      '

            outDome = dicout['Dome'][i]
            outRec = dicout['Rec'][i]

            if '---------------' in outRec :
                continue
                # methode curieuse dans le stat.info sopac : on note
                # l'installation de l'antenne sans le rec : useless

            outRec = receptor_gins_corrector(outRec)[:11]
            outAnt = outAnt[:11]
            void='                     '

            formtuple=(idligne,grg,stat,outRec,Xant,sigoffset,Yant,sigoffset,Zant,sigoffset,outAnt,outDome,void,outStart,outEnd,refchange)

            #outputline ="%s %4s %4s %20s %5.3f %s %5.3f %s %5.3f %s %15s %s %s %s %s" \
            #%(idligne,grg,stat,outRec,Xant,sigoffset,Yant,sigoffset,Zant,sigoffset,outAnt,outDome,outStart,outEnd,refchange)

            outputline=" {0} {1} {2} {3:<11} {4:>-12.3f}  {5:5.3f} {6:>-12.3f}  {7:5.3f} {8:>-12.3f}  {9:5.3f}   {10:15} {11:4} {12} {13} {14} {15}".format(*formtuple)

            outputline=outputline+'\n'
            statgins_filobj.write(outputline)


        statgins_filobj.write('\n')

    if station_info_columns_type == "ulr":
        log.info("NOTA : input station.info is a 'ulr' type, the antenna/reciever type has to be corrected manually in the returned GINS file")


    return outfile

    #oceload_associed(outfile)


def stat_file_GINS_new_fmt(file_out_path,
                           STAT="STAT",
                           xyz=[0.,0.,0.],
                           ecc_une=[0.,0.,0.],
                           rec = "NONE",
                           ant = "NONE",
                           radom = "NONE",
                           ant_id = "0000"):

    F = open(file_out_path,"w+")

    Ellipsoid = """#VERSION 1.0
#RADIUS    0.6378137000000000E+07
#1/FLAT    0.2982572221010000E+03
# ITRF2014"""

    x , y , z = xyz[0] , xyz[1] , xyz[2]
    ecc_u , ecc_n , ecc_e = ecc_une[0] , ecc_une[1] , ecc_une[2]

    l1     = "   99999 M001 {:4}                             mar_xyz {:15.6f} 0.000000 {:15.6f} 0.000000 {:15.6f} 0.000000 vit_xyz  0.00000 0.0000  0.00000 0.0000  0.00000 0.0000  20100101_000000   EURA          i14i14"
    l1_fmt = l1.format(STAT,x,y,z)
    l2 = " 9999999 GRG3 {:4}      {:23}ecc_une {:15.6f} 0.000000 {:15.6f} 0.000000 {:15.6f} 0.000000            {:16}{:4} {:25}20000101_000000                 rinex"
    l2_fmt = l2.format(STAT,rec,ecc_u,ecc_n,ecc_e,ant,radom,ant_id)

    OUT = "\n".join((Ellipsoid , l1_fmt , l2_fmt))

    F.write(OUT)

    F.close()

    return file_out_path



def write_station_info_from_datalists(period_lis_lis,
                                      site_lis,
                                      location_lis,
                                      station_info_out_path):
    """
    

    Parameters
    ----------
    period_lis_lis : list of lists
        list of lists of periods.
    site_lis : list
        list of sites.
    location_lis : list
        list of locations.
    station_info_out_path : str
        the full path (directory+name) of the generated station.info.
        
    Returns
    -------
    None.
    
    Note
    ----
    The different data lists (i.e. period_lis_lis , stat_lis , loc_lis) 
    are produced by:
        * mono_logsheet_read
              do not forget to activate return_lists = True
          
        * multi_logsheet_read 
              do not forget to activate return_dico = False
    """


    si_file = open(station_info_out_path,'w')
    si_file.write('*SITE  Station Name      Session Start      Session Stop       Ant Ht   HtCod  Ant N    Ant E    Receiver Type         Vers                  SwVer  Receiver SN           Antenna Type     Dome   Antenna SN          \n')
    # writing the period in a station info
    for period_lis,sit,loc in zip(period_lis_lis,site_lis,location_lis):
        for period in period_lis:
            d1 , d2 , a , r = period
            # simplifiy name, removing space and non ascii char. ,
            # and fill with spaces to have 16 chars
            name = loc.City_or_Town[0:16].replace(' ','_')
            name = ''.join([i if ord(i) < 128 else '_' for i in name])
            if len(name) != 16:
                name = name.ljust(16)
            stat = sit.Four_Character_ID
            #managing the 4 chars
            stat = stat[0:4]
            if len(stat) != 4:
                stat = stat.ljust(4)

            strtup = stat , name , \
            d1.year,conv.dt2doy(d1),int(d1.hour),int(d1.minute),int(d1.second), \
            d2.year,conv.dt2doy(d2),int(d2.hour),int(d2.minute),int(d2.second), \
            float(a.Up_Ecc) , a.ARPSmart() , float(a.North_Ecc) , float(a.East_Ecc) , \
            r.Receiver_Type , r.Firmware_Version , r.FirmwareSmart() , \
            str(r.Serial_Number)[:20] , a.AntTypSmart() , \
            a.Antenna_Radome_Type , str(a.Serial_Number)[:20]

            outline =  " {}  {:16}  {} {} {:0>2d} {:0>2d} {:0>2d}  {} {} {:0>2d} {:0>2d} {:0>2d}  {:+.4f}  {}  {:+.4f}  {:+.4f}  {:20}  {:<20}  {:>5.2f}  {:<20}  {:15}  {:5}  {}\n".format(*strtup)
            si_file.write(outline)
    si_file.close()
    return None


def write_lfile_from_datalists(site_lis,location_lis,lfile_out_path):
    ''' datalists (period_lis_lis , stat_lis , loc_lis  ) are produced by :
        * mono_logsheet_read
        * multi_logsheet_read  '''
    lf_file = open(lfile_out_path,'w')

    
    proto_str = " {}_GPS {:.4f} {:.4f} {:.4f} {:+6.4f} {:+6.4f} {:+6.4f} {:6.1f} {:5.3f} {:5.3f} {:5.3f} {:6.4f} {:6.4f} {:6.4f}\n"
    
    for sit,loc in zip(site_lis,location_lis):
        vx  , vy  , vz  = 0,0,0
        sx  , sy  , sz  = 0,0,0
        svx , svy , svz = 0,0,0
        
        yr = conv.dt2year_decimal(loc.Reference_epoch)

        outline = proto_str.format(sit.Four_Character_ID ,
                                   loc.X_coordinate_m ,
                                   loc.Y_coordinate_m ,
                                   loc.Z_coordinate_m,
                                   vx  , vy  , vz,
                                   yr  ,
                                   sx  , sy  , sz,
                                   svx , svy , svz)
        
        lf_file.write(outline)
    lf_file.close()
    return None

def header_from_ellipsoid(ellipsoid):
    if ellipsoid in ( 'GRS80' , 'WGS84'):
        header=textwrap.dedent("""APRES ITRF2008 TOUTES STATIONS       ,  R= 6378137.00,   f= 298.257222101
 station type site                    X (m)   sig.        Y (m)   sig.        Z(m)    sig.    Xp (m/an)      Yp (m/an)      Zp (m/an)    date  plaque source
   -9999 M000 Earth c. of mass        0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                              010100
   -1998 EGB  Bias       (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                              010198 311298
    -199 EGB  Bias       (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                              010199 310199
     -61 EGC1 1/year Cos (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                                            dy ecc
     -51 EGS1 1/year Sin (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                                            dy ecc
     -62 EGC2 2/year Cos (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                                            dy ecc
     -52 EGS2 2/year Sin (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                                            dy ecc

   14001 M004 ZIMM              4331297.063 ~0.001   567555.878 ~0.001  4633133.936 ~0.001 -0.0135 ~.0001  0.0181 ~.0000  0.0121 ~.0001 010105        i08i08
 1400101 GRG3 ZIMM TR 4000SSE         0.000  0.000        0.000  0.000        0.000  0.000   TRM14532.00     NONE 3311A                 010593 050897 igs06
 1400102 GRG3 ZIMM TR 4000SSI         0.000  0.000        0.000  0.000        0.000  0.000   TRM14532.00     NONE 3311A                 060897 051198 igs06

   14001 M004 ZIMM              4331297.062 ~0.001   567555.882 ~0.001  4633133.935 ~0.001 -0.0135 ~.0001  0.0181 ~.0000  0.0121 ~.0001 010105        i08i08
 1400106 GRG3 ZIMM TR 4000SSI         0.000  0.000        0.000  0.000        0.000  0.000   TRM29659.00     NONE 99390                 061198 110803 igs06
 1400107 GRG3 ZIMM TR 4700            0.000  0.000        0.000  0.000        0.000  0.000   TRM29659.00     NONE 99390                 120803 210206 igs06
 1400108 GRG3 ZIMM TR NETRS           0.000  0.000        0.000  0.000        0.000  0.000   TRM29659.00     NONE 99390                 220206        igs06

            """)
    elif ellipsoid in ('GRIM' , 'EIGEN'):
        header=textwrap.dedent("""APRES ITRF2008 TOUTES STATIONS       ,  R= 6378136.46,   f= 298.257650
 station type site                    X (m)   sig.        Y (m)   sig.        Z(m)    sig.    Xp (m/an)      Yp (m/an)      Zp (m/an)    date  plaque source
   -9999 M000 Earth c. of mass        0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                              010100
   -1998 EGB  Bias       (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                              010198 311298
    -199 EGB  Bias       (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                              010199 310199
     -61 EGC1 1/year Cos (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                                            dy ecc
     -51 EGS1 1/year Sin (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                                            dy ecc
     -62 EGC2 2/year Cos (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                                            dy ecc
     -52 EGS2 2/year Sin (mm)         0.000 ~0.000        0.000 ~0.000        0.000 ~0.000                                                            dy ecc

   14001 M004 ZIMM              4331297.063 ~0.001   567555.878 ~0.001  4633133.936 ~0.001 -0.0135 ~.0001  0.0181 ~.0000  0.0121 ~.0001 010105        i08i08
 1400101 GRG3 ZIMM TR 4000SSE         0.000  0.000        0.000  0.000        0.000  0.000   TRM14532.00     NONE 3311A                 010593 050897 igs06
 1400102 GRG3 ZIMM TR 4000SSI         0.000  0.000        0.000  0.000        0.000  0.000   TRM14532.00     NONE 3311A                 060897 051198 igs06

   14001 M004 ZIMM              4331297.062 ~0.001   567555.882 ~0.001  4633133.935 ~0.001 -0.0135 ~.0001  0.0181 ~.0000  0.0121 ~.0001 010105        i08i08
 1400106 GRG3 ZIMM TR 4000SSI         0.000  0.000        0.000  0.000        0.000  0.000   TRM29659.00     NONE 99390                 061198 110803 igs06
 1400107 GRG3 ZIMM TR 4700            0.000  0.000        0.000  0.000        0.000  0.000   TRM29659.00     NONE 99390                 120803 210206 igs06
 1400108 GRG3 ZIMM TR NETRS           0.000  0.000        0.000  0.000        0.000  0.000   TRM29659.00     NONE 99390                 220206        igs06

            """)
    else:
        log.error('Check ellipsoid Name')
        return None
    return header

def write_station_file_gins_from_datalists(period_lis_lis,site_lis,location_lis,
                                           station_info_out_path,ellipsoid="GRS80"):
    ''' datalists (period_lis_lis , stat_lis , loc_lis  ) are produced by :
        * mono_logsheet_read
        * multi_logsheet_read  '''

    # Info de referentiel de la 1ère ligne
    refiel = 'i08i08'
    # Info de referentiel des lignes de réoccupations
    refchange = 'igs10'
    # code 'GRG3'
    grg = 'GRG3'
    # sigma sur les offsets
    sigoffset = 0.000
    # sigma sur les positions
    sigposi = 0.030
    # sigma sur les vitesses
    sigvit=0.005
    # Vitesses

    # ID DOMES du pays
    # ex : 971 pour la Guadeloupe
    # mettre 999 dans la plupart des cas
    # ATTENTION : string entre quotes
    idprefix='971'
    idprefix='990'

    # id initial, les stations auront un numero atribué de maniere décroissante
    # 97198 97197 97196 ...
    idstatinit=99

    void='             '
    void2='                     '

    statgins_filobj= open(station_info_out_path,'w+')

    header  = header_from_ellipsoid(ellipsoid)
    statgins_filobj.write(header)

    i = 0
    for i , ( period_lis , site , loca) in enumerate(zip(period_lis_lis,site_lis,location_lis)):
        T = loca.Reference_epoch.strftime('%d%m%y')
        idstatA = site.IERS_DOMES_Number[0:5]
        idstatB = site.IERS_DOMES_Number[5:9]

        # case where an auto DOMES must be given
        if int(idstatA) == 99999:
            idstatA = str(int(idstatA) - i)

        stat = site.Four_Character_ID
        # managing the 4 chars
        stat = stat[0:4]
        if len(stat) != 4:
            stat = stat.ljust(4)

        X = float(loca.X_coordinate_m)
        Y = float(loca.Y_coordinate_m)
        Z = float(loca.Z_coordinate_m)

        vX = float(loca.X_velocity)
        vY = float(loca.Y_velocity)
        vZ = float(loca.Z_velocity)

        sigposi = float(loca.X_coordinate_sigma)
        sigvit  = float(loca.X_velocity_sigma)

        fortupsta=(idstatA,idstatB,stat,void,X,sigposi,Y,Z,vX,sigvit,vY,vZ,T,refiel)
        outputline="   {0} {1} {2}{3}{4:>-12.3f} ~{5:5.3f} {6:>-12.3f} ~{5:5.3f} {7:>-12.3f} ~{5:5.3f} {8:>-7.4f} ~{9} {10:>-7.4f} ~{9} {11:>-7.4f} ~{9} {12}        {13}".format(*fortupsta)

        outputline=outputline+'\n'
        statgins_filobj.write(outputline)

        for j,peri in enumerate(period_lis):
            if j == 0:
                d_new_peri = 0
            else:
                d_new_peri = 1
            start , end , ant ,rec = peri
            idligne = idstatA + str(j).zfill(2)
            outRec = receptor_gins_corrector( rec.Receiver_Type )

            Xant,Yant,Zant = conv.ENU2XYZ_legacy(ant.East_Ecc,ant.North_Ecc,ant.Up_Ecc,X,Y,Z)

            sigoffset = 0.
            outAnt = ant.AntTypSmart()
            outDome = ant.Antenna_Radome_Type[0:4].ljust(4)
            outStart = (start + dt.timedelta(days=d_new_peri)).strftime('%d%m%y')
            if end == dt.datetime(2099,1,1,tzinfo=dateutil.tz.tzutc()):
                outEnd = '      '
            else:
                outEnd = end.strftime('%d%m%y')

            formtuple=(idligne,grg,stat,outRec,Xant,sigoffset,Yant,sigoffset,Zant,sigoffset,outAnt,outDome,void2,outStart,outEnd,refchange)

            outputline=" {0} {1} {2} {3:<11} {4:>-12.3f}  {5:5.3f} {6:>-12.3f}  {7:5.3f} {8:>-12.3f}  {9:5.3f}   {10:15} {11} {12} {13} {14} {15}".format(*formtuple)
            outputline=outputline+'\n'
            statgins_filobj.write(outputline)

        statgins_filobj.write('\n')
    return None

def smart_elt_list(list_raw,n_elt,replacement=''):
    try:
        return list_raw[n_elt]
    except:
        return replacement

def read_rinex_2_dataobjts(rinex_path):

    if utils.empty_file_check(rinex_path):
        log.error('the RINEX file is empty ...')
        log.error(rinex_path)

        return None , None , None , None

    ant_raw   = utils.grep(rinex_path,'ANT #',True)
    rec_raw   = utils.grep(rinex_path,'REC #',True)
    xyz_raw   = utils.grep(rinex_path,'APPROX POSITION XYZ',True).split()
    stat_raw  = utils.grep(rinex_path,'MARKER NAME',True).split()
    domes_raw = utils.grep(rinex_path,'MARKER NUMBER',True).split()
    d_hen_raw = utils.grep(rinex_path,'ANTENNA: DELTA H/E/N',True).split()
    t_raw     = utils.grep(rinex_path,'TIME OF FIRST OBS',True).split()

    Antobj, Recobj , Siteobj , Locobj = read_logsheets.Antenna(),read_logsheets.Receiver(),read_logsheets.Site(),read_logsheets.Location()

    Locobj.X_coordinate_m = float(smart_elt_list(xyz_raw,0))
    Locobj.Y_coordinate_m = float(smart_elt_list(xyz_raw,1))
    Locobj.Z_coordinate_m = float(smart_elt_list(xyz_raw,2))

    Antobj.Up_Ecc    = float(smart_elt_list(d_hen_raw,0))
    Antobj.East_Ecc  = float(smart_elt_list(d_hen_raw,1))
    Antobj.North_Ecc = float(smart_elt_list(d_hen_raw,2))

    Siteobj.Four_Character_ID = smart_elt_list(stat_raw,0)

    if re.search('[0-9]{5}[A-Z][0-9]{3}',smart_elt_list(domes_raw,0)):
        Siteobj.IERS_DOMES_Number =  smart_elt_list(domes_raw,0)
    else:
        Siteobj.IERS_DOMES_Number = str(np.random.randint(99999)).zfill(5) + 'M001' # smart_elt_list(domes_raw,0) A CHANGER !!!!!!!

    Antobj.Antenna_Radome_Type = ant_raw[36:40].strip()
    Antobj.Antenna_Type        = ant_raw[20:36].strip()

    Recobj.Receiver_Type = rec_raw[20:40].strip()

    # ======== OBSOLETE car utilisation de la fct rinex_start_end ========
    ##securité pour les secondes
    #if not (0 <= t_raw[5] < 60):
        #if t_raw[5] > 30:
            #t_raw[5] = 59
        #else:
            #t_raw[5] = 0

    #Date_Installed = dt.datetime(*[int(float(e)) for e in t_raw[0:6]],
                                   #tzinfo=dateutil.tz.tzutc())
    #Date_Removed = Date_Installed + dt.timedelta(days=1)


    Date_Installed , Date_Removed = operational.rinex_start_end(rinex_path,
                                                                 add_tzinfo=1,
                                                                 verbose=0)


    Antobj.Date_Removed = Date_Removed
    Antobj.Date_Installed = Date_Installed

    Recobj.Date_Removed = Date_Removed
    Recobj.Date_Installed = Date_Installed

    Locobj.Reference_epoch = Date_Installed

    return Antobj , Recobj , Siteobj , Locobj


def write_station_file_gins_from_rinex(rinex_path,station_file_out,
                                       stat_code_filename_prioritary = True):
    Antobj , Recobj , Siteobj , Locobj = read_rinex_2_dataobjts(rinex_path)

    rnx_name           = os.path.basename(rinex_path)
    stat_code_filename = os.path.basename(rinex_path)[0:4]

    if Siteobj.Four_Character_ID.upper() != stat_code_filename.upper():
        print('WARN : different 4-char. code in RINEX header and filename')
        print('       for' , rnx_name , '(' , Siteobj.Four_Character_ID , ')')
        if stat_code_filename_prioritary:
            print('       keeping the 4-char. code in filename')
            Siteobj.Four_Character_ID = stat_code_filename.upper()


    write_station_file_gins_from_datalists([[(Antobj.Date_Installed, \
    Antobj.Date_Removed + dt.timedelta(days=1),Antobj,Recobj)]],[Siteobj],[Locobj],station_file_out)

    return None



#                     _   _   _ __  __ ______             _____  _____
#                    | | | \ | |  \/  |  ____|   /\      / ____|/ ____|   /\
#  _ __ ___  __ _  __| | |  \| | \  / | |__     /  \    | |  __| |  __   /  \
# | '__/ _ \/ _` |/ _` | | . ` | |\/| |  __|   / /\ \   | | |_ | | |_ | / /\ \
# | | |  __/ (_| | (_| | | |\  | |  | | |____ / ____ \  | |__| | |__| |/ ____ \
# |_|  \___|\__,_|\__,_| |_| \_|_|  |_|______/_/    \_\  \_____|\_____/_/    \_\
#

def read_nmea(file_path , outtype = "FLH" , df_out = True, use_altitude = False,
              startdate = dt.datetime(1980,1,1) ,
              enddate = dt.datetime(2099,1,1) ,
              export_path = ''):
    """ if export_path != '', export a Matlab readable file to this path
        WARNING !!! la coord de ref est codée en dur, à coriger !!!!"""
    T    = []
    Lat  = []
    Long = []
    Haut = []
    Qual = []
    day = 999
    month = 999
    year = 999
    
    if use_altitude:
        field_h_alt = 9
    else:
        field_h_alt = 11
        
    
    ## open the file as binary because can be mixed with bin data
    for l in open(file_path,"rb"):
        #print(l)
        try:
            l = (l).decode('ascii')
            #print(l)

        except:
            continue
        
        if not l[0] == '$':
            continue
        if 'GPZDA' in l or 'GNZDA' in l:
            f = l.split(',')
            day = int(f[2])
            month = int(f[3])
            year = int(f[4])
        if 'GPGGA' in l or 'GNGGA' in l:
            f = l.split(',')
            h = int(f[1][0:2])
            m = int(f[1][2:4])
            s = int(f[1][4:6])
            qual = int(f[6])
            if  f[5] == 'W':
                EW = -1
            elif f[5] == 'E':
                EW = 1
            else:
                print('WARN: problems....')
            if day == 999:
                continue
            if h == 0 and m == 0 and s == 0:
                continue # c'est sale de zapper minuit mais bon ...
            t = dt.datetime(year,month,day,h,m,s)
            if not ( startdate <= t <= enddate):
                continue
            T.append(t)
            Lat.append(float(f[2][0:2]) + float(f[2][2:]) / 60.)
            Long.append( EW * (float(f[4][0:3]) + float(f[4][3:]) / 60.))
            Haut.append(float(f[field_h_alt]))
            Qual.append(qual)

    X,Y,Z = conv.GEO2XYZ(Lat,Long,Haut)
    f,l,h = np.mean(Lat),np.mean(Long),np.mean(Haut)
    x0,y0,z0 = conv.GEO2XYZ(f,l,h)

    f,l,h =   43.4417981389 , 7.83481522597 , 6.59449264956
    x0,y0,z0 = 4595047.79934 , 632288.017869 , 4363273.52335
    E,N,U = conv.XYZ2ENU_2(X,Y,Z,x0,y0,z0)

    if export_path != '':
        outf = open(export_path,'w+')
        outf.write(' '.join(('#lat0,long0,h0 :',str(f),str(l),str(h),'\n')))
        outf.write(' '.join(('#x0 ,y0 , z0   :',str(x0),str(y0),str(z0),'\n')))
        for i in range(len(T)):
            datalis = [conv.dt2posix(T[i]),T[i].year,T[i].month,T[i].day,
                       T[i].hour,T[i].minute,T[i].second,Lat[i],Long[i],
                        Haut[i],X[i],Y[i],Z[i],E[i],N[i],U[i] ]
            datalis = [str(e) for e in datalis] + ['\n']
            outf.write(','.join(datalis)) #)
        outf.close()
        
    if outtype == 'ENU':
        OUTTUP = T,E,N,U,Qual
        colnames = ('T','E','N','U','Qual')
    elif outtype == 'XYZ':
        OUTTUP = T,X,Y,Z,Qual
        colnames = ('T','X','Y','Z','Qual')        
    else:
        OUTTUP = T,Lat,Long,Haut,Qual
        colnames = ('T','F','L','H','Qual')
        
    if df_out:
        DFout = pd.DataFrame(np.column_stack(OUTTUP))
        DFout.columns = colnames
        DFout = DFout.infer_objects()
        return DFout
    else:
        return OUTTUP
        

#strtd = dt.datetime(2015,06,19,16,00)
#T,E,N,U,Q = read_nmea(file_path='/home/psakicki/geodesea_nav_final.dat',enuout=1,export_path='/home/psakicki/Documents/geodesea_nav_matrix.dat') #,startdate=strtd)

def plot_nmea(T,E,N,U,Q):
    """ a quick and dirty fct to plot the NMEA"""
    plt.close('all')

    y_formatter = matplotlib.ticker.ScalarFormatter(useOffset=1)

    plt.figure()
    plt.plot(T,U,'+')
    plt.title('U')
    ax = plt.gca()
    ax.yaxis.set_major_formatter(y_formatter)

    plt.figure()
    plt.plot(T,N,'+')
    plt.title('N')
    ax = plt.gca()
    ax.yaxis.set_major_formatter(y_formatter)

    plt.figure()
    plt.title('E')
    col = ['r'] * len(E)
    plt.scatter(T,E,Q,marker='+')
    ax = plt.gca()
    ax.yaxis.set_major_formatter(y_formatter)

    plt.figure()
    plt.title('E-N')
    plt.plot(E,N,'+')
    ax = plt.gca()
    ax.yaxis.set_major_formatter(y_formatter)

    plt.figure()
    plt.plot(T,Q,'+')
    ax = plt.gca()
    ax.yaxis.set_major_formatter(y_formatter)

def write_latlontime_file_4_OTPS_tide(outfilepath , lat , lon ,
                                      strt , end = None, sec_step=1 ,
                                      generate_inp_file = True,
                                      tide_model_path = './DATA/Model_atlas',
                                      set_relative_path_in_inp_file=True):
    """
    

    Parameters
    ----------
    outfilepath : TYPE
        DESCRIPTION.
    lat : TYPE
        DESCRIPTION.
    lon : TYPE
        DESCRIPTION.
    strt : datetime
        start as a datetime.
    end : TYPE, optional
        length of the period in sec (integer)
        OR
        if None, strt is a list. The default is None.
    sec_step : TYPE, optional
        DESCRIPTION. The default is 1.
    generate_inp_file : TYPE, optional
        DESCRIPTION. The default is True.
    tide_model_path : TYPE, optional
        DESCRIPTION. The default is './DATA/Model_atlas'.
    set_relative_path_in_inp_file : TYPE, optional
        DESCRIPTION. The default is True.

    Returns
    -------
    outdir : TYPE
        DESCRIPTION.
        
        
    Note
    ----
    lat>0 - degrees North, lon>0 - degrees East
    lat<0 - degrees South, lon<0 - degrees West
    
    lat/lon can be a list, with the same size as dates_lis
    
    """
    
    

    fil = open(outfilepath,'w+')


    if not end:
        dates_lis = strt
        
    elif type(end) is dt.datetime:
        dates_lis  = [strt]
        while dates_lis[-1] <= end:
            dates_lis.append( dates_lis[-1] + dt.timedelta(seconds=sec_step))

    else:
        dates_lis = [strt + dt.timedelta(seconds=x) for x in np.arange(0,end+1,sec_step)]


    if not utils.is_iterable(lat):
        lat = [float(lat)] * len(dates_lis)
        lon = [float(lon)] * len(dates_lis)
    else:
        if len(lat) != len(dates_lis) or len(lon) != len(dates_lis):
            print("ERR : lat/lon are list, but not with the same size as the dates list, please check")
            raise Exception
            

    print('INFO : first & last dates')
    print(dates_lis[0] , dates_lis[-1])

    for d,lat_i,lon_i in zip(dates_lis,lat,lon):
        y  = d.year
        m  = d.month
        da = d.day
        h  = d.hour
        mi = d.minute
        s  = d.second
        lin = '{:12.8}{:12.8}{:7}{:7}{:7}{:7}{:7}{:7}\n'.format(lat_i,lon_i,y,m,da,h,mi,s)
        fil.write(lin)

    if generate_inp_file:
        outdir   = os.path.dirname(outfilepath)
        outname  = os.path.basename(outfilepath) + '.inp'
        outoutname = os.path.basename(outfilepath) + '.out'

        fil2 = open(os.path.join(outdir,outname),'w+')

        fil2.write(tide_model_path + '\n' )
        if not set_relative_path_in_inp_file:
            fil2.write(outfilepath + ' \n')
        else:
            fil2.write("./" + os.path.basename(outfilepath) + ' \n')
            
        fil2.write('z'         + ' \n')
        fil2.write(''          + ' \n')
        fil2.write('AP'        + ' \n')
        fil2.write('geo'       + ' \n')
        fil2.write('1'         + ' \n')
        if not set_relative_path_in_inp_file:
            fil2.write( os.path.join(outdir,outoutname)  + ' \n')
        else:
            fil2.write("./" + outoutname + ' \n')

        fil2.close()

    return outdir

def read_OTPS_tide_file(pathin,print_line=False,return_array=True):
    """
    Return :
        latlis , lonlis , datlis , hlis
    """
    filout = open(pathin)

    latlis , lonlis , datlis , hlis = [],[],[],[]

    for i,l in enumerate(filout):
        f = l.split()
        if np.mod(i , 1000) == 0 and print_line:
            print('line', i)
        try:
            latlis.append(float(f[0]))
            lonlis.append(float(f[1]))
            datlis.append(dateutil.parser.parse(f[2] + ' ' + f[3]))
            hlis.append(float(f[4]))
        except:
            continue
    if return_array:
        return np.array(latlis) , np.array(lonlis) ,  np.array(datlis) , np.array(hlis)
    else:
        return latlis , lonlis , datlis , hlis

def read_snx_trop(snxfile,dataframe_output=True):
    """
    Read troposphere solutions from Troposphere SINEX
    """
    
    STAT , epoc = [] , []
    tro , stro , tgn , stgn , tge , stge = [] , [] , [] , [] , [] , []
    
    flagtrop = False
    
    for line in open(snxfile,"r",encoding = "ISO-8859-1"):
        
        if re.compile('TROP/SOLUTION').search(line):
            flagtrop = True
            continue
        
        if re.compile('-TROP/SOLUTION').search(line):
            flagtrop = False
            continue
        
        if line[0] != ' ':
            continue
        else:
            fields = line.split()
        
        if flagtrop ==True:
            
            STAT.append(fields[0].upper())
            if not ':' in fields[1]:
                epoc.append(conv.convert_partial_year(fields[1]))
            else:
                date_elts_lis = fields[1].split(':')
                yy =  int(date_elts_lis[0]) + 2000
                doy = int(date_elts_lis[1])
                sec = int(date_elts_lis[2])

                epoc.append(conv.doy2dt(yy,doy,seconds=sec))
            
            tro.append(float(fields[2]))
            stro.append(float(fields[3]))
            tgn.append(float(fields[4]))
            stgn.append(float(fields[5]))
            tge.append(float(fields[6]))
            stge.append(float(fields[7]))
            
    outtuple = \
    list(zip(*sorted(zip(STAT , epoc , tro , stro , tgn , stgn , tge , stge))))
    
    if dataframe_output:
        return Tropsinex_DataFrame(outtuple)
                
def Tropsinex_DataFrame(read_sinex_result):
     DF_Sinex = pd.DataFrame.from_records(list(read_sinex_result)).transpose()
     colnam = ['STAT','epoc','tro','stro','tgn','stgn','tge','stge']
     DF_Sinex.columns = colnam
     cols_numeric = ['tro','stro','tgn','stgn','tge','stge']
     DF_Sinex[cols_numeric] = DF_Sinex[cols_numeric].apply(pd.to_numeric, errors='coerce')
     
     return DF_Sinex


def read_sinex(sinex_path_in,
               keep_sites_as_index=False,
               drop_nan=False):
    """
    Read a *coordinate* SINEX file as DataFrame

    Parameters
    ----------
    sinex_path_in : str
        path of the SINEX.
    keep_sites_as_index : Bool, optional
        use the site names as index. The default is False.
    drop_nan : bool or str, optional
        remove the NaN in the DataFrame.
        False: do not remove them
        'all': remove the row if all values are NaN
        'any':  remove the row if at least one value is NaN (dangerous)
        The default is False.

    Returns
    -------
    DFout : DataFrame
        SINEX DataFrame.

    """
    ## Read the blocs
    DFcoor = read_sinex_versatile(sinex_path_in,"SOLUTION/ESTIMATE",
                                           header_line_idx=None)
    DFepoc = read_sinex_versatile(sinex_path_in,"SOLUTION/EPOCHS",
                                           header_line_idx=None)
    
    
    ## Rename the Epoch DF
    DFepoc.rename(columns={0: "STAT",
                           1: "pt",
                           2: "soln",
                           3: "?",
                           4: "start",
                           5: "end",
                           6: "mean"},inplace=True)
    DFepoc.set_index(['STAT','pt','soln'],inplace=True)
        
    ### Rearange the coords DF
    Index = DFcoor[[2,3,4]].drop_duplicates()
    
    Stk_DFcoor2_line = []
    
    for (stat,pt,soln) in zip(Index[2],Index[3],Index[4]):   
        DF2 = DFcoor[(DFcoor[2] == stat) & (DFcoor[3] == pt) & (DFcoor[4] == soln)]
    
        dicttmp = dict()
    
        epoc = DF2[5].unique()[0]    
        dicttmp['epoc'] = epoc
        dicttmp['STAT'] = stat
        dicttmp['pt']   = pt
        dicttmp['soln'] = soln
    
        for typ,comp in itertools.product(('STA','VEL'),
                                          ('X','Y','Z')):
             
            typcomp = typ+comp
            try:
                val = DF2[DF2[1] == typcomp][8].values[0]
                sig = DF2[DF2[1] == typcomp][9].values[0]
            except:
                val,sig = np.nan,np.nan
            
            if typ == 'VEL':
                typdic = 'v'
            else:
                typdic = ''
            
            dicttmp[typdic + comp.lower()] = val
            dicttmp['s' + typdic + comp.lower()] = sig
            
        DFcoor2_line = pd.DataFrame(dicttmp,index=[0])
        Stk_DFcoor2_line.append(DFcoor2_line)
        
    DFcoor2 = pd.concat(Stk_DFcoor2_line)   
    DFcoor2.set_index(['STAT','pt','soln'],inplace=True)
    
    
    ### Concatenate both
    DFout = pd.concat((DFepoc,DFcoor2),axis=1)
        
    ### remove NaN (might be more epochs than coords...)
    if drop_nan:
        DFout.dropna(inplace=True,how=drop_nan)
    
    if not keep_sites_as_index:
        DFout.reset_index(inplace=True)    

    return DFout


def read_sinex_legacy(snxfile,dataframe_output=True):
    """
    This function is depreciated !!!!
    """

    STAT , soln , epoc, AC = [] , [] , [], []
    x , y , z , sx , sy , sz = [] , [] , [] , [] , [] , []
    vx , vy , vz , svx , svy , svz = [] , [] , [] , [] , [] , []
    
    start , end = [] , []

    flagxyz    = False
    flagepochs = False

    for line in open(snxfile,"r",encoding = "ISO-8859-1"):

        if re.compile('SOLUTION/ESTIMATE').search(line):
            flagxyz = not flagxyz
            continue
        
        if re.compile('SOLUTION/EPOCHS').search(line):
            flagepochs = not flagepochs
            continue        

        if line[0] != ' ':
            continue
        else:
            fields = line.split()


        if flagxyz == True:

            if fields[1] == 'STAX':
                split_sp3_name = os.path.basename(snxfile)
                AC.append(split_sp3_name[0:3])
                STAT.append(fields[2].upper())
                soln.append(fields[4])
                if not ':' in fields[5]:
                    epoc.append(conv.convert_partial_year(fields[5]))
                else:
                    date_elts_lis = fields[5].split(':')
                    yy =  int(date_elts_lis[0]) + 2000
                    doy = int(date_elts_lis[1])
                    sec = int(date_elts_lis[2])

                    epoc.append(conv.doy2dt(yy,doy,seconds=sec))

                x.append(float(fields[8]))
                sx.append(float(fields[9]))

            if fields[1] == 'STAY':
                y.append(float(fields[8]))
                sy.append(float(fields[9]))

            if fields[1] == 'STAZ':
                z.append(float(fields[8]))
                sz.append(float(fields[9]))


            if fields[1] == 'VELX':
                vx.append(float(fields[8]))
                svx.append(float(fields[9]))

            if fields[1] == 'VELY':
                vy.append(float(fields[8]))
                svy.append(float(fields[9]))

            if fields[1] == 'VELZ':
                vz.append(float(fields[8]))
                svz.append(float(fields[9]))
                
        if flagepochs:
            start_val = conv.datestr_sinex_2_dt(fields[4])
            end_val   = conv.datestr_sinex_2_dt(fields[5])
            start.append(start_val)
            end.append(end_val)
    
    if not vx:
        nanlist = [np.nan] * len(x)
        vx , vy , vz , svx , svy , svz = list(nanlist) , list(nanlist) , list(nanlist) , \
                                         list(nanlist) , list(nanlist) ,list(nanlist)
                                         

    outtuple = \
    list(zip(*sorted(zip(AC, STAT , soln , epoc , x , y , z , sx , sy , sz ,
                         vx , vy , vz , svx , svy , svz , start , end))))

    if dataframe_output:
        return sinex_DataFrame(outtuple)
    else:
        STAT , soln , epoc , x , y , z , sx , sy , sz , vx , vy , vz , svx , svy , svz , start , end= outtuple
        print('TIPS : this output can be converted directly to a Pandas DataFrame using sinex_DataFrame function')
        return STAT , soln , epoc , x , y , z , sx , sy , sz , vx , vy , vz , svx , svy , svz , start ,  end




def sinex_DataFrame(read_sinex_result):
    DF_Sinex = pd.DataFrame.from_records(list(read_sinex_result)).transpose()
    colnam = ['AC', 'STAT' , 'soln' , 'epoc' , 'x' , 'y' , 'z' , 'sx' , 'sy' , 'sz' , 'vx' , 'vy' , 'vz' , 'svx' , 'svy' , 'svz' , 'start' , 'end']
    DF_Sinex.columns = colnam
    
    cols_numeric = ["x","y","z",
                "sx","sy","sz",
                "vx","vy","vz",
                "svx","svy","svz"]
        
    DF_Sinex[cols_numeric] = DF_Sinex[cols_numeric].apply(pd.to_numeric, errors='coerce')
        

    return DF_Sinex
    
def read_sinex_versatile(sinex_path_in , id_block,
                         convert_date_2_dt = True,
                         header_line_idx = -1,
                         improved_header_detection=True,
                         verbose=True):
    """
    Read a block from a SINEX and return the data as a DataFrame

    Parameters
    ----------
    sinex_path_in : str
        Description param1

    id_block : str
        Name of the block (without "+" or "-")
        
    convert_date_2_dt : bool
        Try to convert a SINEX formated date as a python datetime
        
    header_line_idx : int or None
        If the block header contains several lines, use this line index
        Per default, the last (-1)
        For the first line, use 0
        If no header is properly defined, use None
        
    improved_header_detection : bool
        Improved header detection.
        Works for most cases but sometime the simple version works better.
        (advenced usage)
        Default is True
    
    verbose : bool
        print the header and its field size
        Default is True        
                
    Returns
    -------
    DF : Pandas DataFrame
        Returned DataFrame
    """
    
    ### remove the + or - if any    
    if id_block in ("+","-"):
        id_block = id_block[1:]
    
    id_block_strt = "\+" + id_block
    id_block_end  = "\-" + id_block
    
    Lines_list = utils.extract_text_between_elements_2(sinex_path_in,
                                                       id_block_strt,
                                                       id_block_end)
    Lines_list = Lines_list[1:-1]
    
    if not Lines_list:
        print("ERR : read_sinex_versatile : no block found, ",id_block)
    
    #### Remove commented lines
    Lines_list_header = []
    Lines_list_OK     = []
    header_lines = True
    for i_l , l in enumerate(Lines_list):
        if not l[0] in (" ","\n") and header_lines:
            ## here we store the 1st commented lines i.e. the header
            Lines_list_header.append(l)
        elif l[0] in (" ","\n"):
            ## here we store the data lines (not commented)
            header_lines = False
            Lines_list_OK.append(l)
        else:
            ## here we skip the final commented lines (can happend) 
            continue
                                
    if len(Lines_list_header) > 0 and header_line_idx:
        ### define the header
        header_line = Lines_list_header[header_line_idx]

        Header_split = header_line.split()
        if not improved_header_detection: 
            ### Simple case when the columns are splitted with only a single
            Fields_size = [len(e)+1 for e in Header_split]
        else: 
            ### Smarter case : we search for the n spaces after the column name
            Fields_size = []
            
            
            for fld_head_split in Header_split:
                if fld_head_split[0] == "*": 
                    ## test to manage a * as 1st character
                    ## which can be wrongly interpreted in the regex
                    fld_head_regex = re.compile("\*" + fld_head_split[1:] + " *") 
                else:
                    fld_head_regex_str = fld_head_split + " *"
                    ### update 202110: now the brackets must be escaped (normal btw...)
                    fld_head_regex_str = fld_head_regex_str.replace("[","\[")
                    fld_head_regex_str = fld_head_regex_str.replace("]","\]")
                    fld_head_regex_str = fld_head_regex_str.replace("(","\(")
                    fld_head_regex_str = fld_head_regex_str.replace(")","\)")

                    fld_head_regex = re.compile(fld_head_regex_str) 

                fld_head_space = fld_head_regex.search(header_line)

                Fields_size.append(len(fld_head_space.group()))
                #print(fld_head_space.group())                
                
                # # weak method (210216) archived for legacy
                # fld_head_regex = re.compile(fld_head_split[1:] + " *") #trick:
                # #1st char is removed, because it can be a *
                # #and then screw the regex. This char is re-added at the end
                # #when the len is stored (the "+1" below)
                # fld_head_space = fld_head_regex.search(header_line)
                # Fields_size.append(len(fld_head_space.group()) + 1)
                # ### !!!!! something is weird here !!!!!
                # print(fld_head_space.group())
                # ### and you will see a bug !!!
                # ### PS 210216
        
        if verbose:
            print("INFO : read_sinex_versatile : Auto detected column names/sizes")
            print("**** Raw header line in the file:")
            print(header_line)
            print("**** Splited header for the DataFrame:")
            print(Header_split)
            print("**** Size of the fields")
            print(Fields_size)
    
        ### Add the header in the big string 
        Lines_str_w_head = header_line + "".join(Lines_list_OK)

        ### Read the file
        try:
            DF = pd.read_fwf(StringIO(Lines_str_w_head),widths=Fields_size)
        except pd.errors.EmptyDataError as ee:
            print("ERR: something goes wrong in the header index position")
            print("     try to give its right position manually with header_line_idx")
            raise(ee)

        DF.set_axis(Header_split, axis=1, inplace=True)
        
        ### Rename the 1st column (remove the comment marker)
        DF.rename(columns={DF.columns[0]:DF.columns[0][1:]}, inplace=True)

    else: # no header in the SINEX
        Lines_str = "".join(Lines_list_OK)
        DF = pd.read_csv(StringIO(Lines_str),header=None ,
                             delim_whitespace=True)

    regex_time = "(([0-9]{2}|[0-9]{4}):[0-9]{3}|[0-9]{7}):[0-9]{5}"
    for col in DF.columns:
        if convert_date_2_dt and re.match(regex_time,
                                          str(DF[col].iloc[0])):
            try:
                DF[col] = DF[col].apply(lambda x : conv.datestr_sinex_2_dt(x))
            except Exception as e:
                print("WARN : read_sinex_versatile : convert date string to datetime failed")
                print(e)
                pass
        
    return DF

def read_sinex_bench_antenna(sinex_in):
    F = open(sinex_in,"r")

    line_good_stk = []
    for l in F:
        if l[0] == " ":
            line_good_stk.append(l)

    T = StringIO("".join(line_good_stk))

    header_cols = ["TYP" ,"STA_" ,"Occ", "Code","PT", "T",
                   "___x/up___" ,"___y/n____", "___z/e____" ,
                   "_Data_Start", "_Data_End__",
                   "__Antenna_type______" ,"Radome","__S/N__"]
    DFantenna = pd.read_table(T,delim_whitespace = True,error_bad_lines=False,names=header_cols)

    return DFantenna


def read_rinex_nav(fn,writeh5=None,version=2):
    """
    Based on Michael Hirsch readRinexNav function
    Manage RINEX3
    http://gage14.upc.es/gLAB/HTML/GPS_Navigation_Rinex_v2.11.html
    """

    with open(fn , 'r') as f:
        #find end of header, which has non-constant length
        while True:

            line = f.readline()

            if 'RINEX VERSION' in line:
                if float(line.split()[0]) < 3:
                    version = 2
                else:
                    version = 3


            if 'END OF HEADER' in line: break

            if version == 2:
                startcol = 3 #column where numerical data starts
                nfloat=19 #number of text elements per float data number
                nline=7 #number of lines per record
            elif version == 3:
                startcol = 5 #column where numerical data starts
                nfloat=19 #number of text elements per float data number
                nline=7 #number of lines per record


        #handle frame by frame
        sv = []; epoch=[]; raws=''


        while True:
            headln = f.readline()
            if not headln: break
            #handle the header
            if version == 2:
                sv.append(headln[:2])
                year = int(headln[2:5])
                if 80<= year <=99:
                    year+=1900
                elif year<80: #good till year 2180
                    year+=2000

                epoch.append(dt.datetime(year =year,
                                      month   =int(headln[5:8]),
                                      day     =int(headln[8:11]),
                                      hour    =int(headln[11:14]),
                                      minute  =int(headln[14:17]),
                                      second  =int(headln[17:20]),
                                      microsecond=int(headln[21])*100000))


                raw = (headln[22:].rstrip() +
                       ''.join(f.readline()[startcol:].rstrip() for _ in range(nline-1))
                       +f.readline()[startcol:79].rstrip())




            if version == 3:
                fields = headln[:23].split(' ')
                sv.append(fields[0])
                year = int(fields[1])

                epoch.append(dt.datetime(year = year,
                                      month   =int(fields[2]),
                                      day     =int(fields[3]),
                                      hour    =int(fields[4]),
                                      minute  =int(fields[5]),
                                      second  =int(fields[6])))

                """
                now get the data.
                Use rstrip() to chomp newlines consistently on Windows and Python 2.7/3.4
                Specifically [:-1] doesn't work consistently as .rstrip() does here.
                """
                raw = (headln[23:].rstrip() + ' ' +
                       ''.join(f.readline()[startcol:].rstrip() + ' ' for _ in range(nline-1))
                       +f.readline()[startcol:80].rstrip())

            raws += raw + '\n'

    raws = raws.replace('D','E')

    strio = BytesIO(raws.encode())
    darr = np.genfromtxt(strio,delimiter=nfloat)

    nav= pd.DataFrame(darr, epoch,
               ['SVclockBias','SVclockDrift','SVclockDriftRate','IODE',
                'Crs','DeltaN','M0','Cuc','Eccentricity','Cus','sqrtA','TimeEph',
                'Cic','OMEGA','CIS','Io','Crc','omega','OMEGA DOT','IDOT',
                'CodesL2','GPSWeek','L2Pflag','SVacc','SVhealth','TGD','IODC',
                'TransTime','FitIntvl',"spare1","spare2"])

    if version == 3:
        Const = [e[0]  for e in sv]
        SV    = [int(e[1:]) for e in sv]
        nav['Const'] = pd.Series(np.array(Const), index=nav.index)
        nav['sv'] = pd.Series(np.array(SV), index=nav.index)
    elif version == 2:
        rinexnav_type = os.path.basename(fn)[-1]
        if rinexnav_type == 'n':
            nav['Const'] = pd.Series(['G']*len(nav.index), index=nav.index)
        else:
            nav['Const'] = pd.Series([rinexnav_type.upper()]*len(nav.index), index=nav.index)
        nav['sv'] = pd.Series(np.array(sv), index=nav.index)



    if False:
        h5fn = fn.with_suffix('.h5')
        print(('saving NAV data to {}'.format(h5fn)))
        nav.to_hdf(h5fn,key='NAV',mode='a',complevel=6,append=False)

    return nav


def unzip_gz_Z(inp_gzip_file,out_gzip_file='',remove_inp=False, force = False):
    """
    if out_gzip_file if not precised, file will be extracted in the same folder
    as inp_gzip_file

    .Z decompression is implemented, but is very unstable (avoid .Z, prefer .gz)
    """

    import gzip

    if inp_gzip_file.endswith('.gz'):
        is_gz = True
    elif inp_gzip_file.endswith('.Z'):
        is_gz = False
    else:
        is_gz = True

    if out_gzip_file == '':
        out_gzip_file = os.path.join(os.path.dirname(inp_gzip_file) ,
                                     '.'.join(os.path.basename(inp_gzip_file).split('.')[:-1]))


    if os.path.isfile(out_gzip_file) and not force:
        print('INFO : ' , out_gzip_file , 'already exists, skiping (use force option)')
        pass
    else:
        if is_gz:
            with gzip.open(inp_gzip_file, 'rb') as f_in, open(out_gzip_file, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
        else:
            print("WARN : zlib decompress is unstable !! and .Z should be definitly avoided ... ")
            #str_object1 = open(inp_gzip_file, 'rb').read()
            #str_object2 = zlib.decompress(str_object1)
            #f = open(out_gzip_file, 'wb')
            #f.write(str_object2)
            #f.close()

            out_gzip_file = utils.uncompress(inp_gzip_file)

        print('INFO : uncompressing ' + inp_gzip_file + " to " + out_gzip_file )


    ### Removing part
    if remove_inp and os.path.getsize(out_gzip_file) > 0:
        print("INFO : removing " + inp_gzip_file)
        os.remove(inp_gzip_file)

    return out_gzip_file


 #  ______                _   _                _____                                         _ 
 # |  ____|              | | (_)              / ____|                                       | |
 # | |__ _   _ _ __   ___| |_ _  ___  _ __   | |  __ _ __ __ ___   _____ _   _  __ _ _ __ __| |
 # |  __| | | | '_ \ / __| __| |/ _ \| '_ \  | | |_ | '__/ _` \ \ / / _ \ | | |/ _` | '__/ _` |
 # | |  | |_| | | | | (__| |_| | (_) | | | | | |__| | | | (_| |\ V /  __/ |_| | (_| | | | (_| |
 # |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|  \_____|_|  \__,_| \_/ \___|\__, |\__,_|_|  \__,_|
 #                                                                        __/ |                
 #                                                                       |___/       



########################################################################'

# ## Read the blocs
# DFcoor = files_rw.read_sinex_versatile(sinex_path_in,"SOLUTION/ESTIMATE",
#                                        header_line_idx=None)
# DFepoc = files_rw.read_sinex_versatile(sinex_path_in,"SOLUTION/EPOCHS",
#                                        header_line_idx=None)

# ## Rename the Coord DF
# DFcoor.rename(columns={0: "INDEX",
#                        1: "TYPE",
#                        2: "CODE",
#                        3: "PT",
#                        4: "SOLN",
#                        5: "REF_EPOCH",
#                        6: "UNIT",
#                        7: "N",
#                        8: "ESTIMATED_VALUE",
#                        9: "STD_DEV"},inplace=True)

# ## Rename the Epoch DF
# DFepoc.rename(columns={0: "CODE",
#                        1: "PT",
#                        2: "SOLN",
#                        3: "T",
#                        4: "DATA_START",
#                        5: "DATA_END",
#                        6: "MEAN_EPOCH"},inplace=True)


# try:
#     DFepoc[DFepoc["SOLN"].str.contains("----")] == 1
# except:
#     pass

# DFcoor = DFcoor[DFcoor["TYPE"].str.contains("STA")]

# AAA = DFcoor.groupby(["CODE",
#                       "SOLN"])
# for a in AAA:
#     print(a[1]["ESTIMATED_VALUE"].T)




# Convert sp3 2 gins
# brouillon
#
#satdic = OrderedDict()
#
#sp3in = "/media/psakicki/AF0E-DA43/CHAL_BUGS/TEMP_DATA/jp211236.sp3"
#orbginsout = '/home/psakicki/aaa_FOURBI/outgins'
#
#for l in open(sp3in):
#    if l[0] in ('#','+','%','/'):
#        continue
#
#    f = l.split()
#
#    if l[0] == '*':
#        ll = [int(float(e)) for e in f[1:]]
#        epoch = dt.datetime(*ll)
#        continue
#    if l[0] == 'P':
#        prn = int(f[0][2:])
#        x = float(f[1]) * 10**3
#        y = float(f[2]) * 10**3
#        z = float(f[3]) * 10**3
#
#        print epoch,prn,x,y,z
#
#        if not satdic.has_key(prn):
#            satdic[prn] = []
#
#        jjul = conv.dt2jjulCNES(epoch)
#        sec = conv.dt2secinday(epoch) + 19
#
#        satdic[prn].append((jjul,sec,x,y,z))
#
#fil = open(orbginsout,'w+')
#
#for k,vv in satdic.iteritems():
#    for v in vv:
#        if k in (11,13,14,18,20,28):
#            finalprn = 66600 + k
#        elif k in (2,15,17,21):
#            finalprn = 88800 + k
#        else:
#            finalprn = 77700 + k
#
#        jjul = v[0]
#        sec  = v[1]
#        x = v[2]
#        y = v[3]
#        z = v[4]
#
#        finalline =  '  {:5} {:6} {:12.6f} tai xyz ine  0 {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:+.14E} {:.3f}  {:.3E}\n'.format(finalprn,jjul,float(sec),0,0,0,0,0,0,0,0,0,x,y,z,0,0,0,0.,0)
#        fil.write(finalline)
#
#fil.close()



#                   ___    _____   ____  __  __ ______  _____                                  _
#                  |__ \  |  __ \ / __ \|  \/  |  ____|/ ____|                                | |
#   ___ _____   __    ) | | |  | | |  | | \  / | |__  | (___    _ __ ___  __ _ _   _  ___  ___| |_
#  / __/ __\ \ / /   / /  | |  | | |  | | |\/| |  __|  \___ \  | '__/ _ \/ _` | | | |/ _ \/ __| __|
# | (__\__ \\ V /   / /_  | |__| | |__| | |  | | |____ ____) | | | |  __/ (_| | |_| |  __/\__ \ |_
#  \___|___/ \_/   |____| |_____/ \____/|_|  |_|______|_____/  |_|  \___|\__, |\__,_|\___||___/\__|
#                                                                           | |
#                                                                           |_|

#def csv2DOMES(csvin):

#filein = '/home/psakicki/gin/TP/GWADA/gps_continu_full.csv'
#import pandas as pd
#df = pd.read_csv(open(filein))
#saved_column = list(df['Nom'])
#
#
#for i,row in df.iterrows():
#    name = eval(row['Nom'])
#    stat = row['Alias']
#    outdomesfile = open('/home/psakicki/gin/TP/GWADA/' + stat + '_DOMES_request.txt' ,'w+' )
#    outdomesfile.write(' 1. Request from (full name) : ' + '\n')
#    outdomesfile.write('Agency                   : '+ 'Institut de Pysique du Globe de Paris' +  '\n')
#    outdomesfile.write('E-mail                   : '+ '\n')
#    outdomesfile.write('Date                     : '+ '\n')
#    outdomesfile.write(' '+ '\n')
#    outdomesfile.write('2. Site Name                : ' + name + '\n')
#    outdomesfile.write('3. Country                  : ' + 'Guadeloupe , French West Indies' + '\n'  )
#    outdomesfile.write('4. Point Description        : ' + row['Type'] + '\n' )
#    outdomesfile.write('5. DOMES Number             : ' + 'N/A' + '\n' )
#    outdomesfile.write('6. Local Number             : ' + 'N/A' + '\n')
#    outdomesfile.write('7. 4-Char Code              : ' + row['Alias'] + '\n' )
#    outdomesfile.write(' '+ '\n')
#    outdomesfile.write('8. Approximate Position ' + '\n')
#    outdomesfile.write('   Latitude (deg min)       : '  + str(row['Lat. (WGS84)']) + '\n' )
#    outdomesfile.write('   Longitude (deg min)      : '  + str(row['Lon. (WGS84)']) + '\n' )
#    outdomesfile.write('   Elevation (m)            : '  + str(row['Elev. (m)']) + '\n')
#    outdomesfile.write(' '+ '\n')
#    outdomesfile.write('9. Instrument               : '  + str('GPS on ' + row['Type']) + '\n' )
#    outdomesfile.write('10. Date of Installation     : ' + str(row['Début / Installation']) + '\n')
#    outdomesfile.write(' '+ '\n')
#    outdomesfile.write('11. Operation Contact Name   : ' + '\n')
#    outdomesfile.write('    Agency                   : ' + '\n')
#    outdomesfile.write('    E-mail                   : ' + '\n')
#    outdomesfile.write(' ' + '\n')
#    outdomesfile.write('12. Site Contact Name        : ' + '\n')
#    outdomesfile.write('    Agency                   : ' + '\n')
#    outdomesfile.write('    E-mail                   : ' + '\n')


#  ________   ________ __  __ _____  _      ______  _____
# |  ____\ \ / /  ____|  \/  |  __ \| |    |  ____|/ ____|
# | |__   \ V /| |__  | \  / | |__) | |    | |__  | (___
# |  __|   > < |  __| | |\/| |  ___/| |    |  __|  \___ \
# | |____ / . \| |____| |  | | |    | |____| |____ ____) |
# |______/_/ \_\______|_|  |_|_|    |______|______|_____/
#
#
## CATS
#
#neulist = ['/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wABMF_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wADE0_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wASF0_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wCBE0_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wDHS0_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wDSD0_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wFNA0_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wFSDC_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wHOUE_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wLAM0_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wMGL0_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wPDB0_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wPSA1_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wSOUF_GINS_PS.neu',
#'/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wTDB0_GINS_PS.neu']
#
#statinfo_path = '/home/psakicki/THESE/CODES/stationinfo2gins/station.info.ovsg'
#
#for neu in neulist:
#    print neu
#    statinfo_2_cats(statinfo_path,neu)
#
#read_station_info_solo(statinfo_path,'ABMF')

#read_station_info_solo('/home/psakicki/THESE/CODES/stationinfo2gins/station.info.ovsg','HOUE')
#statname_of_catsfile('/home/psakicki/gin/TP/GWADA/DATA_CATS/NEU/wASF0_GINS_PS.neu')



# GINS
#station_info_2_gins('/home/psakicki/gin/TP/GWADA/station.info.ovsg.volobsis','/home/psakicki/gin/TP/GWADA/lfile_mod_mk2','/home/psakicki/gin/TP/GWADA/GWADA_mk2.stat')
#coordfile = '/home/psakicki/gin/TP/GWADA/RINX2/KARIB/pbo.snaps_igs08.vel'
#statinfo = '/home/psakicki/gin/TP/GWADA/RINX2/KARIB/station.info'

#specific = ['CN40' , 'CN19' , 'CN38' , 'SAN0' , 'CN35' , 'ROA0' , 'CN29' , 'CN18' , 'CBMD' , 'LCSB' , 'GCFS' , 'GCEA' , 'CN10' , 'CN12' , 'PRCG' , 'SMRT' , 'BARA']

#station_info_2_gins(statinfo,coordfile,'/home/psakicki/gin/TP/GWADA/KARIB_mk3.stat',specific_stats_lis=specific)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  9 09:23:54 2019

@author: Chaiyaporn Kitpracha
"""

#from geodezyx import *                   # Import the GeodeZYX modules
#from geodezyx.externlib import *         # Import the external modules
#from geodezyx.megalib.megalib import *   # Import the legacy modules names

import numpy as np
import re
import pandas as df
from geodezyx import conv,utils

def PWV_conversion(zwd,Tm):
    """
    This function convert from Zenith Wet delay to Precipitate Water Vapor (PWV)

    Parameters
    ----------
    zwd :
        Zenith wet delay in meters
    Tm:
        Mean temperature of troposphere

    Returns
    ----------
    PWV :
        Precipitate Water Vapor in mm

    Sources
    ----------
    Solution and Constant k2' k3 from Atmospheric effects in Space Geodesy Chapter 3.
    """
    if utils.is_iterable(zwd):
        PWV = []
        for zwd_m,Tm_m in zip(zwd,Tm):
            pwv = PWV_conversion(zwd_m,Tm_m)
            PWV.append(pwv)
        return np.array(PWV)

    else:
        k1 = 77.689
        k2 = 71.2952
        Md = 28.965 / 100 # g/mol -> kg/mol
        Mw = 18.016 / 100 # g/mol -> kg/mol
        k2d = k2 - (k1*(Mw/Md))
        k3 = 375463
        R = 8.3143
        Rhow = 999.97

        CF = 10e6 * Mw / (Rhow * R * (k2d + (k3/Tm)))
        PWV = CF* zwd * 1000
        return np.round(PWV,2)

def Tm_bevis(Ts):
    """
    This function determines mean temperature based on surface temperature using Bevis equation

    Parameters
    ----------
    Ts:
        Surface temperature in Kelvin
    Returns
    ----------
    Tm:
        Mean temperature in Kevlin
    Sources
    ----------
    Atmospheric effects in Space Geodesy Chapter 3.
    """

    Tm = 70.2 + 0.72*Ts
    return np.round(Tm,2)

def trop_saast(p,dlat,hell,t=0,e=0,mode="dry"):
    """
    This subroutine determines the zenith total delay based on the equation by Saastamoinen (1972) as refined by Davis et al. (1985)

    Parameters
    ----------
    p :
        pressure in hPa

    dlat :
        ellipsoidal latitude in radians

    t :
         temperature in Celcius

    e :
         water vapor pressure in hPa

    hell :
         ellipsoidal height in meters

    mode :
         dry, wet or total

    Returns
    ----------
    res :
        zenith total delay in m (depend on mode)

    Source
    ----------
     c Reference:
     Saastamoinen, J., Atmospheric correction for the troposphere and
     stratosphere in radio ranging of satellites. The use of artificial
     satellites for geodesy, Geophys. Monogr. Ser. 15, Amer. Geophys. Union,
     pp. 274-251, 1972.
     Davis, J.L, T.A. Herring, I.I. Shapiro, A.E.E. Rogers, and G. Elgered,
     Geodesy by Radio Interferometry: Effects of Atmospheric Modeling Errors
     on Estimates of Baseline Length, Radio Science, Vol. 20, No. 6,
     pp. 1593-1607, 1985.

    """
    if utils.is_iterable(dlat):
        ztd = []
        for p_m,dlat_m,hell_m,t_m,e_m in zip(p,dlat,hell,t,e):
            ztd_m = trop_saast(p_m,dlat_m,hell_m,t_m,e_m,mode)
            ztd.append(ztd_m)
        return np.array(ztd)
    else:
        f = 1-0.00266*np.cos(2*dlat) - 0.00000028*hell # calculate denominator f
        t = t + 273.15 #convert celcius to kelvin
        if mode == "dry":
        	res = 0.0022768*p/f
        elif mode == "wet":
        	res = (0.22768e-2) * ((0.1255e+4)+(t+0.5e-1)) * (e/t)
        elif mode == "total":
        	zhd = 0.0022768*p/f
        	zwd = (0.22768e-2) * ((0.1255e+4)+(t+0.5e-1)) * (e/t)
        	res = zhd + zwd
    
        return np.round(res,4)

def read_grid_gpt(grid_name,cols=64):
    grid = np.loadtxt(grid_name,usecols=range(cols),comments="%",dtype=float)
    return grid

def calc_stand_ties_gpt3(epoc, lat_ref , lon_ref , h_ref , lat_rov , lon_rov , h_rov ,grid,unit="mm"):
    """
    Determine standard atmospheric ties from meteological information from GPT3 with analytical equation from Teke et al. (2011)

    Parameters:
    ----------
        epoc :
            time in Python datetime
        lat_ref :
            Latitude of Ref. station
        lat_rov :
            Latitude of Rov. station
        h_ref :
            Height of Ref. station
        h_rov :
            Height of Rov. station
        grid_file :
            meteological grid file
        unit :
            in meters (m) or milimeters (mm)

    Return:
    ----------
        ties :
            Standard ties of total delay in milimeters or meters

    """
    if utils.is_iterable(lat_ref):
        ties = []
        for epoc_m,lat_ref_m,lon_ref_m,h_ref_m,lat_rov_m,lon_rov_m,h_rov_m in zip(epoc,lat_ref,lon_ref,h_ref,lat_rov,lon_rov,h_rov):
            ties_m = calc_stand_ties_gpt3(epoc_m,lat_ref_m,lon_ref_m,h_ref_m,lat_rov_m,lon_rov_m,h_rov_m,grid)
            ties.append(ties_m)
        return np.array(ties)

    else:
        met_ref = gpt3(epoc,lat_ref,lon_ref,h_ref,grid)

        p0 , t0 , e0  = met_ref[0] , met_ref[1] , met_ref[4]
        t0 = t0 + 273.15 #convert to Kelvin

        p = p0 * ((1 - (0.0065 * (h_rov-h_ref)  / t0))  ** (9.80665 / ( 0.0065 * 287.058 )))
        dZHD = (0.0022768*(p0 - p)) / ( 1 - 0.00266*np.cos(2 * lat_ref) - 0.28e-6*h_ref)
        dZWD = 2.789*e0 / (t0**2)  * (( 5383 / t0) - 0.7803) * 0.0065 * (h_rov - h_ref)

        if unit == "m":
            return np.round(dZHD+dZWD,4)

        elif unit == "mm":
            return np.round((dZHD+dZWD) * 1000,4)

def calc_stand_ties(epoc, lat_ref, h_ref, h_rov, p0, t0, e0 ,unit="mm"):
    """
    Determine standard atmospheric ties with analytical equation from Teke et al. (2011)

    Parameters:
    ----------
        epoc :
            time in Python datetime
        lat_ref :
            Latitude of Ref. station
        h_ref :
            Height of Ref. station
        h_rov :
            Height of Rov. station
        p0:
            Pressure of Ref. station in hPa
        t0:
            Temperature in C of Ref. station
        e0:
            Water vapor pressure of Ref. station in hPa
        unit :
            in meters (m) or milimeters (mm)

    Return:
    ----------
        ties :
            Standard ties of total delay in milimeters or meters

    """
    if utils.is_iterable(lat_ref):
        ties = []
        for epoc_m,lat_ref_m,h_ref_m,h_rov_m,p0_m,t0_m,e0_m in zip(epoc, lat_ref, h_ref, h_rov, p0, t0, e0):
            ties_m = calc_stand_ties(epoc_m,lat_ref_m,h_ref_m,h_rov_m,p0_m,t0_m,e0_m)
            ties.append(ties_m)
        return np.array(ties)

    else:
        t0 = t0 + 273.15 #convert to Kelvin

        p = p0 * ((1 - (0.0065 * (h_rov-h_ref)  / t0))  ** (9.80665 / ( 0.0065 * 287.058 )))
        dZHD = (0.0022768*(p0 - p)) / ( 1 - 0.00266*np.cos(2 * lat_ref) - 0.28e-6*h_ref)
        dZWD = 2.789*e0 / (t0**2)  * (( 5383 / t0) - 0.7803) * 0.0065 * (h_rov - h_ref)

        if unit == "m":
            return np.round(dZHD+dZWD,4)

        elif unit == "mm":
            return np.round((dZHD+dZWD) * 1000,4)

def gpt3(dtin,lat,lon,h_ell,C,it=0):


    """
    This subroutine determines pressure, temperature, temperature lapse rate,
    mean temperature of the water vapor, water vapour pressure, hydrostatic
    and wet mapping function coefficients ah and aw, water vapour decrease
    factor, geoid undulation and empirical tropospheric gradients for
    specific sites near the earth's surface.
    It is based on a 5 x 5 degree external grid file ('gpt3_5.grd') with mean
    values as well as sine and cosine amplitudes for the annual and
    semiannual variation of the coefficients.

    Parameters:
    ----------
    dtin :
        datatime in Python datetime object

    lat:
        ellipsoidal latitude in radians [-pi/2:+pi/2]

    lon:
        longitude in radians [-pi:pi] or [0:2pi]

    h_ell:
        ellipsoidal height in m

    it:
        case 1 no time variation but static quantities, case 0 with time variation (annual and semiannual terms)

    Returns:
    ----------
    p:
        pressure in hPa

    T:
        temperature in degrees Celsius

    dT:
        temperature lapse rate in degrees per km

    Tm:
        mean temperature weighted with the water vapor in degrees Kelvin

    e:
        water vapour pressure in hPa

    ah:
        hydrostatic mapping function coefficient at zero height (VMF3)

    aw:
        wet mapping function coefficient (VMF3)

    la:
        water vapour decrease factor

    undu:
        geoid undulation in m

    Gn_h:
        hydrostatic north gradient in m

    Ge_h:
        hydrostatic east gradient in m

    Gn_w:
        wet north gradient in m

    Ge_w:
        wet east gradient in m

    Notes
    ----------
        Modified for Python by Chaiyaporn Kitpracha

    Source
    ----------
        (c) Department of Geodesy and Geoinformation, Vienna University of
        Technology, 2017

        The copyright in this document is vested in the Department of Geodesy and
        Geoinformation (GEO), Vienna University of Technology, Austria. This document
        may only be reproduced in whole or in part, or stored in a retrieval
        system, or transmitted in any form, or by any means electronic,
        mechanical, photocopying or otherwise, either with the prior permission
        of GEO or in accordance with the terms of ESTEC Contract No.
        4000107329/12/NL/LvH.

        D. Landskron, J. Böhm (2018), VMF3/GPT3: Refined Discrete and Empirical Troposphere Mapping Functions,
        J Geod (2018) 92: 349., doi: 10.1007/s00190-017-1066-2.
        Download at: https://link.springer.com/content/pdf/10.1007%2Fs00190-017-1066-2.pdf
    """
    lat = np.array(lat)
    lon = np.array(lon)
    h_ell = np.array(h_ell)
    # Extract data from grid
    p_grid    = C[:,2:7]          # pressure in Pascal
    T_grid    = C[:,7:12]         # temperature in Kelvin
    Q_grid    = C[:,12:17]/1000   # specific humidity in kg/kg
    dT_grid   = C[:,17:22]/1000   # temperature lapse rate in Kelvin/m
    u_grid    = C[:,22]           # geoid undulation in m
    Hs_grid   = C[:,23]           # orthometric grid height in m
    ah_grid   = C[:,24:29]/1000   # hydrostatic mapping function coefficient, dimensionless
    aw_grid   = C[:,29:34]/1000   # wet mapping function coefficient, dimensionless
    la_grid   = C[:,34:39]           # water vapor decrease factor, dimensionless
    Tm_grid   = C[:,39:44]        # mean temperature in Kelvin
    Gn_h_grid = C[:,44:49]/100000    # hydrostatic north gradient in m
    Ge_h_grid = C[:,49:54]/100000    # hydrostatic east gradient in m
    Gn_w_grid = C[:,54:59]/100000    # wet north gradient in m
    Ge_w_grid = C[:,59:64]/100000    # wet east gradient in m

    # Convert from datetime to doy
    doy = float(conv.dt2doy(dtin)) + conv.dt2fracday(dtin)

    # determine the GPT3 coefficients

    # mean gravity in m/s**2
    gm = 9.80665
    # molar mass of dry air in kg/mol
    dMtr = 28.965e-3
    # universal gas constant in J/K/mol
    Rg = 8.3143

    # factors for amplitudes
    if it==1: # then  constant parameters
        cosfy = 0
        coshy = 0
        sinfy = 0
        sinhy = 0
    else:
        cosfy = np.cos(doy/365.25*2*np.pi)   # coefficient for A1
        coshy = np.cos(doy/365.25*4*np.pi)   # coefficient for B1
        sinfy = np.sin(doy/365.25*2*np.pi)   # coefficient for A2
        sinhy = np.sin(doy/365.25*4*np.pi)   # coefficient for B2

    nstat = lat.size

    # initialization
    p    = np.zeros([nstat , 1])
    T    = np.zeros([nstat , 1])
    dT   = np.zeros([nstat , 1])
    Tm   = np.zeros([nstat , 1])
    e    = np.zeros([nstat , 1])
    ah   = np.zeros([nstat , 1])
    aw   = np.zeros([nstat , 1])
    la   = np.zeros([nstat , 1])
    undu = np.zeros([nstat , 1])
    Gn_h = np.zeros([nstat , 1])
    Ge_h = np.zeros([nstat , 1])
    Gn_w = np.zeros([nstat , 1])
    Ge_w = np.zeros([nstat , 1])

    if lon < 0:
        plon = (lon + 2*np.pi)*180/np.pi
    else:
        plon = lon*180/np.pi

    ppod = (-lat + np.pi/2)*180/np.pi

    ipod = np.floor(ppod+1)
    ilon = np.floor(plon+1)

    # changed for the 1 degree grid
    diffpod = (ppod - (ipod - 0.5))
    difflon = (plon - (ilon - 0.5))

    if ipod == 181:
        ipod = 180

    if ilon == 361:
        ilon = 1

    if ilon == 0:
        ilon = 360

    indx = np.zeros(4)
    indx[0] = (ipod - 1)*360 + ilon

    # near the poles: nearest neighbour interpolation, otherwise: bilinear
    # with the 1 degree grid the limits are lower and upper
    bilinear = 0
    if ppod > 0.5 and ppod < 179.5:
           bilinear = 1



    if bilinear == 0:
        ix = int(indx[0]) - 1

        # transforming ellipsoidal height to orthometric height
        undu = u_grid[ix]
        hgt = h_ell-undu

        # pressure, temperature at the height of the grid
        T0 = T_grid[ix,0] + T_grid[ix,1]*cosfy + T_grid[ix,2]*sinfy + T_grid[ix,3]*coshy + T_grid[ix,4]*sinhy
        p0 = p_grid[ix,0] + p_grid[ix,1]*cosfy + p_grid[ix,2]*sinfy + p_grid[ix,3]*coshy + p_grid[ix,4]*sinhy

        # specific humidity
        Q = Q_grid[ix,0] + Q_grid[ix,1]*cosfy + Q_grid[ix,2]*sinfy + Q_grid[ix,3]*coshy + Q_grid[ix,4]*sinhy

        # lapse rate of the temperature
        dT = dT_grid[ix,0] + dT_grid[ix,1]*cosfy + dT_grid[ix,2]*sinfy + dT_grid[ix,3]*coshy + dT_grid[ix,4]*sinhy

        # station height - grid height
        redh = hgt - Hs_grid[ix]

        # temperature at station height in Celsius
        T = T0 + dT*redh - 273.15

        # temperature lapse rate in degrees / km
        dT = dT*1000

        # virtual temperature in Kelvin
        Tv = T0*[1+0.6077*Q]

        c = gm*dMtr/[Rg*Tv]

        # pressure in hPa
        p = [p0*np.exp[-c*redh]]/100


        # hydrostatic and wet coefficients ah and aw
        ah = ah_grid[ix,0] + ah_grid[ix,2]*cosfy + ah_grid[ix,3]*sinfy + ah_grid[ix,4]*coshy + ah_grid[ix,5]*sinhy
        aw = aw_grid[ix,0] + aw_grid[ix,2]*cosfy + aw_grid[ix,3]*sinfy + aw_grid[ix,4]*coshy + aw_grid[ix,5]*sinhy

        # water vapour decrease factor la
        la = la_grid[ix,0] + \
                la_grid[ix,1]*cosfy + la_grid[ix,2]*sinfy + \
                la_grid[ix,3]*coshy + la_grid[ix,4]*sinhy

        # mean temperature Tm
        Tm = Tm_grid[ix,0] + \
                Tm_grid[ix,1]*cosfy + Tm_grid[ix,2]*sinfy + \
                Tm_grid[ix,3]*coshy + Tm_grid[ix,4]*sinhy

        # north and east gradients [total, hydrostatic and wet]
        Gn_h = Gn_h_grid[ix,0] + Gn_h_grid[ix,1]*cosfy + Gn_h_grid[ix,2]*sinfy + Gn_h_grid[ix,3]*coshy + Gn_h_grid[ix,4]*sinhy
        Ge_h = Ge_h_grid[ix,0] + Ge_h_grid[ix,1]*cosfy + Ge_h_grid[ix,2]*sinfy + Ge_h_grid[ix,3]*coshy + Ge_h_grid[ix,4]*sinhy
        Gn_w = Gn_w_grid[ix,0] + Gn_w_grid[ix,1]*cosfy + Gn_w_grid[ix,2]*sinfy + Gn_w_grid[ix,3]*coshy + Gn_w_grid[ix,4]*sinhy
        Ge_w = Ge_w_grid[ix,0] + Ge_w_grid[ix,1]*cosfy + Ge_w_grid[ix,2]*sinfy + Ge_w_grid[ix,3]*coshy + Ge_w_grid[ix,4]*sinhy

        # water vapor pressure in hPa
        e0 = Q*p0/[0.622+0.378*Q]/100 # on the grid
        e = e0*(100*p/p0)**(la+1)   # on the station height - (14] Askne and Nordius, 1987

    else:
        ipod1 = ipod + 1*np.sign(diffpod)
        ilon1 = ilon + 1*np.sign(difflon)

        # changed for the 1 degree grid
        if ilon1 == 361:
            ilon1 = 1

        if ilon1 == 0:
            ilon1 = 360

        # get the number of the line
        # changed for the 1 degree grid
        indx[1] = (ipod1 - 1)*360 + ilon  # along same longitude
        indx[2] = (ipod  - 1)*360 + ilon1 # along same polar distance
        indx[3] = (ipod1 - 1)*360 + ilon1 # diagonal
        indx = indx.astype(int)
        indx = indx - 1

        # transforming ellipsoidal height to orthometric height: Hortho = -N + Hell
        undul = u_grid[indx]
        hgt = h_ell-undul

        # pressure, temperature at the height of the grid
        T0 = T_grid[indx,0] + T_grid[indx,1]*cosfy + T_grid[indx,2]*sinfy + T_grid[indx,3]*coshy + T_grid[indx,4]*sinhy
        p0 = p_grid[indx,0] + p_grid[indx,1]*cosfy + p_grid[indx,2]*sinfy + p_grid[indx,3]*coshy + p_grid[indx,4]*sinhy

        # humidity
        Ql = Q_grid[indx,0] + Q_grid[indx,1]*cosfy + Q_grid[indx,2]*sinfy + Q_grid[indx,3]*coshy + Q_grid[indx,4]*sinhy

        # reduction = stationheight - gridheight
        Hs1 = Hs_grid[indx]
        redh = hgt - Hs1

        # lapse rate of the temperature in degree / m
        dTl = dT_grid[indx,0] + dT_grid[indx,1]*cosfy + dT_grid[indx,2]*sinfy + dT_grid[indx,3]*coshy + dT_grid[indx,4]*sinhy

        # temperature reduction to station height
        Tl = T0 + dTl * redh - 273.15

        # virtual temperature
        Tv = T0 *(1+0.6077*Ql)
        c = gm*dMtr / (Rg*Tv)

        # pressure in hPa
        pl = (p0 *np.exp(-c * redh))/100

        # hydrostatic and wet coefficients ah and aw
        ahl = ah_grid[indx,0] + ah_grid[indx,1]*cosfy + ah_grid[indx,2]*sinfy + ah_grid[indx,3]*coshy + ah_grid[indx,4]*sinhy
        awl = aw_grid[indx,0] + aw_grid[indx,1]*cosfy + aw_grid[indx,2]*sinfy + aw_grid[indx,3]*coshy + aw_grid[indx,4]*sinhy

        # water vapour decrease factor la
        lal = la_grid[indx,0] + la_grid[indx,1]*cosfy + la_grid[indx,2]*sinfy + la_grid[indx,3]*coshy + la_grid[indx,4]*sinhy

        # mean temperature of the water vapor Tm
        Tml = Tm_grid[indx,0] + Tm_grid[indx,1]*cosfy + Tm_grid[indx,2]*sinfy + Tm_grid[indx,3]*coshy + Tm_grid[indx,4]*sinhy

        # north and east gradients [total, hydrostatic and wet]
        Gn_hl = Gn_h_grid[indx,0] + Gn_h_grid[indx,1]*cosfy + Gn_h_grid[indx,2]*sinfy + Gn_h_grid[indx,3]*coshy + Gn_h_grid[indx,4]*sinhy
        Ge_hl = Ge_h_grid[indx,0] + Ge_h_grid[indx,1]*cosfy + Ge_h_grid[indx,2]*sinfy + Ge_h_grid[indx,3]*coshy + Ge_h_grid[indx,4]*sinhy
        Gn_wl = Gn_w_grid[indx,0] + Gn_w_grid[indx,1]*cosfy + Gn_w_grid[indx,2]*sinfy + Gn_w_grid[indx,3]*coshy + Gn_w_grid[indx,4]*sinhy
        Ge_wl = Ge_w_grid[indx,0] + Ge_w_grid[indx,1]*cosfy + Ge_w_grid[indx,2]*sinfy + Ge_w_grid[indx,3]*coshy + Ge_w_grid[indx,4]*sinhy

        # water vapor pressure in hPa
        e0 = Ql *p0 /(0.622+0.378*Ql)/100 # on the grid
        el =  e0 * (100 *pl /p0)**(lal+1)  # on the station height - [14] Askne and Nordius, 1987

        dnpod1 = abs(diffpod) # distance nearer point
        dnpod2 = 1 - dnpod1   # distance to distant point
        dnlon1 = abs(difflon)
        dnlon2 = 1 - dnlon1

        # pressure
        R1 = dnpod2*pl[0]+dnpod1*pl[1]
        R2 = dnpod2*pl[2]+dnpod1*pl[3]
        p = dnlon2*R1+dnlon1*R2

        # temperature
        R1 = dnpod2*Tl[0]+dnpod1*Tl[1]
        R2 = dnpod2*Tl[2]+dnpod1*Tl[3]
        T = dnlon2*R1+dnlon1*R2

        # temperature in degree per km
        R1 = dnpod2*dTl[0]+dnpod1*dTl[1]
        R2 = dnpod2*dTl[2]+dnpod1*dTl[3]
        dT = (dnlon2*R1+dnlon1*R2)*1000

        # water vapor pressure in hPa
        R1 = dnpod2*el[0]+dnpod1*el[1]
        R2 = dnpod2*el[2]+dnpod1*el[3]
        e = dnlon2*R1+dnlon1*R2

        # ah and aw
        R1 = dnpod2*ahl[0]+dnpod1*ahl[1]
        R2 = dnpod2*ahl[2]+dnpod1*ahl[3]
        ah = dnlon2*R1+dnlon1*R2
        R1 = dnpod2*awl[0]+dnpod1*awl[1]
        R2 = dnpod2*awl[2]+dnpod1*awl[3]
        aw = dnlon2*R1+dnlon1*R2

        # undulation
        R1 = dnpod2*undul[0]+dnpod1*undul[1]
        R2 = dnpod2*undul[2]+dnpod1*undul[3]
        undu = dnlon2*R1+dnlon1*R2

        # water vapor decrease factor la
        R1 = dnpod2*lal[0]+dnpod1*lal[1]
        R2 = dnpod2*lal[2]+dnpod1*lal[3]
        la = dnlon2*R1+dnlon1*R2

        # gradients
        R1 = dnpod2*Gn_hl[0]+dnpod1*Gn_hl[1]
        R2 = dnpod2*Gn_hl[2]+dnpod1*Gn_hl[3]
        Gn_h = (dnlon2*R1 + dnlon1*R2)
        R1 = dnpod2*Ge_hl[0]+dnpod1*Ge_hl[1]
        R2 = dnpod2*Ge_hl[2]+dnpod1*Ge_hl[3]
        Ge_h = (dnlon2*R1 + dnlon1*R2)
        R1 = dnpod2*Gn_wl[0]+dnpod1*Gn_wl[1]
        R2 = dnpod2*Gn_wl[2]+dnpod1*Gn_wl[3]
        Gn_w = (dnlon2*R1 + dnlon1*R2)
        R1 = dnpod2*Ge_wl[0]+dnpod1*Ge_wl[1]
        R2 = dnpod2*Ge_wl[2]+dnpod1*Ge_wl[3]
        Ge_w = (dnlon2*R1 + dnlon1*R2)

        # mean temperature of the water vapor Tm
        R1 = dnpod2*Tml[0]+dnpod1*Tml[1]
        R2 = dnpod2*Tml[2]+dnpod1*Tml[3]
        Tm = dnlon2*R1+dnlon1*R2

    soln = [np.round(p,3),np.round(T,3),np.round(dT,3),np.round(Tm,3),np.round(e,3), \
            np.round(ah,3),np.round(aw,3),np.round(la,3),np.round(undu,3),np.round(Gn_h,3),np.round(Ge_h,3), \
            np.round(Gn_w,3),np.round(Ge_w,3)]
    return soln

def gpt2_5(mjd,lat,lon,HELL,IT,VEC):
    """
     (c) Department of Geodesy and Geoinformation, Vienna University of
     Technology, 2013

     The copyright in this document is vested in the Department of Geodesy and
     Geoinformation (GEO), Vienna University of Technology, Austria. This document
     may only be reproduced in whole or in part, or stored in a retrieval
     system, or transmitted in any form, or by any means electronic,
     mechanical, photocopying or otherwise, either with the prior permission
     of GEO or in accordance with the terms of ESTEC Contract No.
     4000107329/12/NL/LvH.
     ---

     This subroutine determines pressure, temperature, temperature lapse rate,
     mean temperature of the water vapor, water vapour pressure, hydrostatic
     and wet mapping function coefficients ah and aw, water vapour decrease
     factor and geoid undulation for specific sites near the Earth surface.
     It is based on a 5 x 5 degree external grid file ('gpt2_5.grd') with mean
     values as well as sine and cosine amplitudes for the annual and
     semiannual variation of the coefficients.

     The hydrostatic mapping function coefficients have to be used with the
     height dependent Vienna Mapping Function 1 (vmf_ht.f) because the
     coefficients refer to zero height.

     Example 1 (Vienna, 2 August 2012, with time variation):

     dmjd = 56141.d0
     dlat(1) = 48.20d0*pi/180.d0
     dlon(1) = 16.37d0*pi/180.d0
     hell(1) = 156.d0
     nstat = 1
     it = 0

     output:
     p = 1002.56 hPa
     T = 22.12 deg Celsius
     dT = -6.53 deg / km
     Tm = 281.11 K
     e = 16.72 hPa
     ah = 0.0012647
     aw = 0.0005726
     la = 2.6964
     undu = 44.06 m

     Example 2 (Vienna, 2 August 2012, without time variation, i.e. constant values):

     dmjd = 56141.d0
     dlat(1) = 48.20d0*pi/180.d0
     dlon(1) = 16.37d0*pi/180.d0
     hell(1) = 156.d0
     nstat = 1
     it = 1

     output:
     p = 1003.49 hPa
     T = 11.95 deg Celsius
     dT = -5.47 deg / km
     Tm = 273.00 K
     e = 10.23 hPa
     ah = 0.0012395
     aw = 0.0005560
     la = 2.6649
     undu = 44.06 m

     Klemens Lagler, 2 August 2012
     Johannes Boehm, 6 August 2012, revision
     Klemens Lagler, 21 August 2012, epoch change to January 1 2000
     Johannes Boehm, 23 August 2012, adding possibility to determine constant field
     Johannes Boehm, 27 December 2012, reference added
     Johannes Boehm, 10 January 2013, correction for dlat = -90 degrees
      (problem found by Changyong He)
     Johannes Boehm, 21 May 2013, bug with dmjd removed (input parameter dmjd was replaced
     unintentionally; problem found by Dennis Ferguson)
     Gregory Pain,   17 June 2013, adding water vapour decrease factor la
     Gregory Pain,   01 July 2013, adding mean temperature Tm
     Gregory Pain,   30 July 2013, changing the method to calculate the water vapor partial pressure (e)
     Gregory Pain,   31 July 2013, correction for (dlat = -90 degrees, dlon = 360 degrees)
     Johannes Boehm, 27 December 2013, copyright notice added
     Johannes Boehm, 25 August 2014, reference changed to Boehm et al. in GPS
     Solutions

    Source
    ----------
         J. Böhm, G. Möller, M. Schindelegger, G. Pain, R. Weber, Development of an
         improved blind model for slant delays in the troposphere (GPT2w),
         GPS Solutions, 2014, doi:10.1007/s10291-014-0403-7
    Notes:
         Modified for Python by: Chaiyaporn Kitpracha

    Parameters:
    ----------
    mjd:
         modified Julian date (scalar, only one epoch per call is possible)

    lat:
         ellipsoidal latitude in degrees

    lon:
         longitude in degrees

    HELL:
         ellipsoidal height in m

    IT:
         case 1: no time variation but static quantities
    case 0: with time variation (annual and semiannual terms)

    VEC:
        GPT2 grid data in numpy array size 5 x 5 degree ('gpt2_5.grd')

    Returns:
    ----------
     p:
         pressure in hPa

     T:
         temperature in degrees Celsius

     dT:
         temperature lapse rate in degrees per km

     e:
         water vapour pressure in hPa

     undu:
         geoid undulation in m

    """
    GM = 9.80665
    DMTR = 28.965e-3
    RG = 8.3143

    DLAT = lat*np.pi/180
    DLON = lon*np.pi/180

    DMJD1= mjd - 51544.5
    if IT==1:
        COSFY = 0;
        COSHY = 0;
        SINFY = 0;
        SINHY = 0;
    else:
        COSFY = np.cos(DMJD1/365.25*2*np.pi)
        COSHY = np.cos(DMJD1/365.25*4*np.pi)
        SINFY = np.sin(DMJD1/365.25*2*np.pi)
        SINHY = np.sin(DMJD1/365.25*4*np.pi)

    PGRID = VEC[0:2592,2:7]
    TGRID = VEC[0:2592,7:12]
    QGRID = VEC[0:2592,12:17] / 1000
    DTGRID = VEC[0:2592,17:22] / 1000
    U = VEC[0:2592,22]
    HS = VEC[0:2592,23]


    if DLON < 0:
        PLON = (DLON+2*np.pi)*180/np.pi
    else:
        PLON = DLON*180/np.pi

    PPOD = (-DLAT + np.pi/2)*180/np.pi

    IPOD = np.floor((PPOD+5)/5)
    ILON = np.floor((PLON+5)/5)

    DIFFPOD = (PPOD - (IPOD*5 - 2.5))/5
    DIFFLON = (PLON - (ILON*5 - 2.5))/5

    if IPOD == 37:
        IPOD = 36

    INDX=np.zeros(4)
    INDX[0] = (IPOD - 1)*72 + ILON
    IBILINEAR = 0

    if PPOD > 2.5 and PPOD < 177.5:
        IBILINEAR =1

    if IBILINEAR == 0:
        UNDU=0.0
        IX = INDX[0] -1
        UNDU = U[IX]
        HGT = HELL-UNDU

        T0 = TGRID[IX,0] + TGRID[IX,1]*COSFY + TGRID[IX,2]*SINFY + TGRID[IX,3]*COSHY + TGRID[IX,4]*SINHY
        P0 = PGRID[IX,0] + PGRID[IX,1]*COSFY + PGRID[IX,2]*SINFY + PGRID[IX,3]*COSHY + PGRID[IX,4]*SINHY
        Q = QGRID[IX,0] + QGRID[IX,1]*COSFY + QGRID[IX,2]*SINFY + QGRID[IX,3]*COSHY + QGRID[IX,4]*SINHY
        DT = DTGRID[IX,0] + DTGRID[IX,1]*COSFY + DTGRID[IX,2]*SINFY + DTGRID[IX,3]*COSHY + DTGRID[IX,4]*SINHY

        REDH = HGT - HS[IX]
        T = T0 + DT*REDH - 273.15
        DT = DT * 1000
        TV = T0 * (1 + 0.6077*Q)
        C = GM * DMTR / (RG*TV)
        P = (P0*np.exp(-C*REDH))/100
        E = (Q*P)/(0.622 + 0.378*Q)
    else:
        IPOD1 = IPOD + 1*np.sign(DIFFPOD)
        ILON1 = ILON + 1*np.sign(DIFFLON)
        if ILON1 == 73:
            ILON1 = 1

        if ILON1 == 0:
            ILON1 = 72

        INDX[1] = (IPOD1 - 1)*72 + ILON;
        INDX[2] = (IPOD  - 1)*72 + ILON1;
        INDX[3] = (IPOD1 - 1)*72 + ILON1;

        INDX = INDX.astype(int)
        UNDUL=np.zeros(4)
        QL = np.zeros(4)
        DTL = np.zeros(4)
        TL = np.zeros(4)
        PL = np.zeros(4)

        INDX = INDX - 1

        for L in range(0,4):
            UNDUL[L] = U[INDX[L]]
            HGT = HELL-UNDUL[L]
            T0 = TGRID[INDX[L],0] + TGRID[INDX[L],1]*COSFY + TGRID[INDX[L],2]*SINFY + TGRID[INDX[L],3]*COSHY + TGRID[INDX[L],4]*SINHY
            P0 = PGRID[INDX[L],0] + PGRID[INDX[L],1]*COSFY + PGRID[INDX[L],2]*SINFY + PGRID[INDX[L],3]*COSHY + PGRID[INDX[L],4]*SINHY
            QL[L] = QGRID[INDX[L],0] + QGRID[INDX[L],1]*COSFY + QGRID[INDX[L],2]*SINFY + QGRID[INDX[L],3]*COSHY + QGRID[INDX[L],4]*SINHY
            HS1 = HS[INDX[L]]
            REDH = HGT - HS1

            DTL[L] = DTGRID[INDX[L],0] + DTGRID[INDX[L],1]*COSFY + DTGRID[INDX[L],2]*SINFY + DTGRID[INDX[L],3]*COSHY + DTGRID[INDX[L],4]*SINHY

            TL[L] = T0 + DTL[L]*REDH - 273.15
            TV = T0*(1+0.6077*QL[L])
            C = GM*DMTR/(RG*TV)
            PL[L] = (P0*np.exp(-C*REDH))/100

        DNPOD1 = np.absolute(DIFFPOD)
        DNPOD2 = 1 - DNPOD1
        DNLON1 = np.absolute(DIFFLON)
        DNLON2 = 1 - DNLON1


        R1 = DNPOD2*PL[0]+DNPOD1*PL[1]
        R2 = DNPOD2*PL[2]+DNPOD1*PL[3]
        P = DNLON2*R1+DNLON1*R2

        R1 = DNPOD2*TL[0]+DNPOD1*TL[1]
        R2 = DNPOD2*TL[2]+DNPOD1*TL[3]
        T = DNLON2*R1+DNLON1*R2

        R1 = DNPOD2*DTL[0]+DNPOD1*DTL[1]
        R2 = DNPOD2*DTL[2]+DNPOD1*DTL[3]
        DT = (DNLON2*R1+DNLON1*R2)*1000

        R1 = DNPOD2*QL[0]+DNPOD1*QL[1]
        R2 = DNPOD2*QL[2]+DNPOD1*QL[3]
        Q = DNLON2*R1+DNLON1*R2
        E = (Q*P)/(0.622+0.378*Q)

        R1 = DNPOD2*UNDUL[0]+DNPOD1*UNDUL[1]
        R2 = DNPOD2*UNDUL[2]+DNPOD1*UNDUL[3]
        UNDU = DNLON2*R1+DNLON1*R2
    return round(P,2),round(T,2),round(DT,2),round(E,2),round(UNDU,2)

def vmf1(ah,aw,dt,dlat,zd):
    """
    This subroutine determines the VMF1 (Vienna Mapping Functions 1) for specific sites.

    Parameters
    ----------
    ah:
        hydrostatic coefficient a
    aw:
        wet coefficient a
    dt:
        datetime in python datetime
    dlat:
        ellipsoidal latitude in radians
    zd:
        zenith distance in radians

    Return
    ----------
    vmf1h:
        hydrostatic mapping function
    vmf1w:
        wet mapping function

    Reference
    ----------
    Boehm, J., B. Werl, H. Schuh (2006), Troposphere mapping functions for GPS and very long baseline interferometry
    from European Centre for Medium-Range Weather Forecasts operational analysis data,
    J. Geoph. Res., Vol. 111, B02406, doi:10.1029/2005JB003629.

    Notes
    ----------
    Written by Johannes Boehm, 2005 October 2

    Translated to python by Chaiyaporn Kitpracha
    """

    pi = 3.14159265359
    dmjd = conv.dt2MJD(dt)
    doy = dmjd  - 44239.0 + 1 - 28

    bh = 0.0029
    c0h = 0.062

    if dlat < 0:
        phh  = pi
        c11h = 0.007
        c10h = 0.002
    else:
         phh  = 0
         c11h = 0.005
         c10h = 0.001

    ch = c0h + ((np.cos(doy/365.25*2*pi + phh)+1)*c11h/2 \
                + c10h)*(1-np.cos(dlat))

    sine = np.sin(pi/2 - zd)
    beta = bh / (sine + ch)
    gamma = ah/ (sine + beta)
    topcon = (1 + ah/(1 + bh/(1 + ch)))
    vmf1h = topcon/(sine+gamma)

    bw = 0.00146
    cw = 0.04391
    beta   = bw/( sine + cw )
    gamma  = aw/( sine + beta)
    topcon = (1 + aw/(1 + bw/(1 + cw)))
    vmf1w = topcon/(sine+gamma)

    return vmf1h,vmf1w



####################### Compare troposphere delay  ####################################################







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.marine contains functions to interpolate 
GEBCO bathymetry.

it can be imported directly with:
from geodezyx import marine

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""


########## BEGIN IMPORT ##########
#### External modules
import numpy as np

#### geodeZYX modules
from geodezyx import utils

#### Import star style
# from geodezyx import *                   # Import the GeodeZYX modules
# from geodezyx.externlib import *         # Import the external modules
# from geodezyx.megalib.megalib import *   # Import the legacy modules names

#### Import the logger
import logging
log = logging.getLogger(__name__)

##########  END IMPORT  ##########

def gebco_bathy_grid_extractor(dataset,latmin,latmax,lonmin,lonmax):
    """
    for safety reasons, lat and lon input MUST BE in the dataset,
    replaced by the closest elsewhere

    return latnew , lonnew , Znew
    """
    
    lon = dataset['lon'][:]
    lat = dataset['lat'][:]
    Z   = dataset['elevation'][:]

    latmin_dec = latmin
    latmax_dec = latmax
    lonmin_dec = lonmin
    lonmax_dec = lonmax

    #latmin_dec = conv.dms2dec_num(*latmin)
    #latmax_dec = conv.dms2dec_num(*latmax)
    #lonmin_dec = conv.dms2dec_num(*lonmin)
    #lonmax_dec = conv.dms2dec_num(*lonmax)


    if not np.any(latmin_dec == lat):
        log.warning("... replacing to the nearest")
        latmin_dec = utils.find_nearest(lat,latmin_dec)[0]

    if not np.any(latmax_dec == lat):
        log.warning("... replacing to the nearest")
        latmax_dec = utils.find_nearest(lat,latmax_dec)[0]

    if not np.any(lonmin_dec == lon):
        log.warning("... replacing to the nearest")
        lonmin_dec = utils.find_nearest(lon,lonmin_dec)[0]

    if not np.any(lonmax_dec == lon):
        log.warning("... replacing to the nearest")
        lonmax_dec = utils.find_nearest(lon,lonmax_dec)[0]

    boollat = np.logical_and(latmin_dec <= lat, lat <= latmax_dec)
    boollon = np.logical_and(lonmin_dec <= lon, lon <= lonmax_dec)


    gridlat = np.tile( boollat[:,np.newaxis]  , (1,len(lon)))
    gridlon = np.tile( boollon , (len(lat),1))

    Znew   = Z[ gridlon * gridlat ]
    latnew = lat[boollat]
    lonnew = lon[boollon]
    Znew   = Znew.reshape(len(latnew),len(lonnew))

    return latnew , lonnew , Znew



import bisect
from collections import defaultdict
from collections import Counter
import copy
import datetime as dt
import fnmatch
from ftplib import FTP
import glob
from io import BytesIO,StringIO
import inspect
import itertools
import linecache
import math
import matplotlib
import matplotlib.pyplot as plt
import multiprocessing as mp
from natsort import natsorted, ns
import numpy as np
import numpy as npaa
import operator
import os 
from os import remove, close
import pandas as df
import pandas as pd
import pickle
import scipy
### desactivated MAYOBS25 230925
#from scipy.signal import butter, lfilter, freqz
import shutil
import string
import struct
import sys
import subprocess
import re
import tempfile
from tempfile import mkstemp
import tabulate
import time
import urllib
import uuid



# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.geodyn contains functions to determine the 
Euler pole of a tectonic plate.

it can be imported directly with:
from geodezyx import geodyn

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox

This module is based on the work of :
Goudarzi, M. A., Cocard, M., & Santerre, R. (2014). 
EPC: Matlab software to estimate Euler pole parameters. GPS Solutions, 18(1), 
153–162. https://doi.org/10.1007/s10291-013-0354-4
"""

import numpy as np

#  ______      _             _____      _      
# |  ____|    | |           |  __ \    | |     
# | |__  _   _| | ___ _ __  | |__) |__ | | ___ 
# |  __|| | | | |/ _ \ '__| |  ___/ _ \| |/ _ \
# | |___| |_| | |  __/ |    | |  | (_) | |  __/
# |______\__,_|_|\___|_|    |_|   \___/|_|\___|
#                                              
    
### Euler Pole determination


def euler_pole_calc(lat_ref,long_ref,vn_ref,ve_ref,
                    incvn_ref=None,incve_ref=None,Rt=6.3710088e6):
    """
    Compute the Euler pole of a set of reference points
    
    Parameters
    ----------
    lat_ref,long_ref : list or numpy.array
        latitude and longitude of the reference points (deg)

    vn_ref,ve_ref : list or numpy.array
        north and east velocities of the reference points (m/yr)
        
    incvn_ref,incve_ref : list or numpy.array
        uncertainties on north and east velocities of the reference points (m/yr)

    Rt : float
        Earth Radius
        IUGG value = 6.3710088e6
        IAU value  = 6.378e6
        
                
    Returns
    -------
    w : numpy.array
        Euler vector (rad/yr)
    
    wratedeg : float
        Rate of rotation (deg/Myr)
    
    wlat,wlong : float
        latitude and longitude of the Euler pole (deg)
        
    wwmat : numpy.array
        weight matrix (for debug)

    desmat : numpy.array
        design matrix (for debug)
        
    nrmatinv : 2-tuple
        output of scipy's lstsq fct (for debug)

    Source
    ------
    based on
    Goudarzi, M. A., Cocard, M., & Santerre, R. (2014). EPC: Matlab software to estimate Euler pole parameters. GPS Solutions, 18(1), 153–162. https://doi.org/10.1007/s10291-013-0354-4
    
    Notes
    -----
    w is the common element for all Euler pole functions
    
    Should remains in rad/yr
    
    Written by C. Geisert (ENSTA/LIENSs) - 2017
    """
    if (incve_ref is None) or (incvn_ref is None):
        incve_ref = np.ones(len(ve_ref))
        incvn_ref = np.ones(len(vn_ref))
            
    all_pos_ref =np.column_stack((lat_ref,long_ref))
    coords=(all_pos_ref*np.pi/180)  # [rad] position des stations de la plaque de ref lat,long [°] > lat,long [rad]
    dm=topo2dm(coords)  # [rad]
    desmat=Rt*(dm) #design matrix [m] [rad]
    Tdesmat=np.transpose(desmat) # [m] [rad]
    n  = len(all_pos_ref)
    vel_vector = list(np.zeros([2 * n,1]))     # vn,ve ; vn, ve alternes  vecteur (vn,ve)0 puis (vn,ve)1, puis stations suivantes..
    for  index in range (0,2*n-1,2):    
        vel_vector[index]   =  vn_ref[int(index/2)]
        vel_vector[index+1]   =   ve_ref[int(index/2)]#vn_ref, ve_ref > vel_vector    
    s=[0]*len(incvn_ref)                              
    covars=np.transpose(np.array( [incvn_ref,incve_ref,s])) #[m/yr]     
    wwmat=covarvec2wtmat(covars) #weight matrix of observations  #[m/yr]
    wwmat=np.array(wwmat[0])
    nrmat=Tdesmat.dot(wwmat).dot(desmat) # [m] [rad] 
    lennrmat=np.shape(nrmat) # []
    nrmatinv=np.linalg.lstsq(nrmat,np.eye(lennrmat[0]),rcond=None)  # [m] [│rad]         
    A=(Tdesmat.dot(wwmat)).dot((vel_vector))  # [m, rad] * [m/yr]   
    w=nrmatinv[0].dot(A) # taille 3x3  # [m, rad] * [m/yr]  #   [rad/yr]
#    wx=w[0]  #[rad/ yr]  
#    wy=w[1]  #[rad/ yr] 
#    wz=w[2]  #[rad/ yr] 
#    wrate=np.sqrt(wx**2+wy**2+wz**2)*1e6    # [rad/yr]> [rad/Myr]
#    wratedeg=wrate*180/np.pi  # [deg/Myr]
#    wlat=(np.arctan(wz/(np.sqrt(wx**2+wy**2))))*180/np.pi # [°] 
#    wlong=(np.arctan(wy/wx))*180/np.pi # [°]     

    wlat,wlong,wrate,wratedeg = euler_pole_vector_to_latlongrate(w)  
    
    return w,wratedeg,wlat,wlong,wwmat,desmat,nrmatinv


def euler_pole_vector_to_latlongrate(w):   
    """
    Convert Euler pole vector to latitude longitude and rate  
        
    Parameters
    ----------
    w : numpy.array
        Pole of rotation vector computed by euler_pole_calc (rad/yr)

    Returns
    -------
    wlat,wlong : float
        latitude and longitude of the Euler pole (deg)
        
    wrate : float
        Rate of rotation (rad/Myr)
        
    wratedeg : float
        Rate of rotation (deg/Myr)
        
    Notes
    -----
    w is the common element for all Euler pole functions
    
    Should remains in rad/yr
    
    Written by P. Sakic based on C. Geisert's work
    """
    wx=w[0]  #[rad/ yr]  
    wy=w[1]  #[rad/ yr] 
    wz=w[2]  #[rad/ yr] 
    wrate=np.sqrt(wx**2+wy**2+wz**2)*1e6    # [rad/yr]> [rad/Myr]
    wratedeg=wrate*180/np.pi  # [deg/Myr]
    wlat=(np.arctan(wz/(np.sqrt(wx**2+wy**2))))*180/np.pi # [°] 
    wlong=(np.arctan(wy/wx))*180/np.pi # [°]   
    return wlat,wlong,wrate,wratedeg
    

def euler_vels_relative_to_ref(w,lat_ITRF,long_ITRF,vn_ITRF,ve_ITRF,
                               incvn_ITRF=None,incve_ITRF=None,Rt=6.378e6):
    """
    Compute relative velocities of points with respect to a reference plate/Euler pole
        
    Parameters
    ----------
    w : numpy.array
        Pole of rotation vector computed by euler_pole_calc (rad/yr)

    lat_ITRF,long_ITRF : list or numpy.array
        latitude and longitude of the points (deg)
        
    vn_ITRF,ve_ITRF : list or numpy.array
        north and east velocities of the points (m/yr)

    incvn_ITRF,incve_ITRF : list or numpy.array
        uncertainties on north and east velocities of the points (m/yr)
                              
    Returns
    -------
    vel_reltoref : numpy.array
        relative velocities of points with respect to the Euler pole w
    
    Notes
    -----
    w is the common element for all Euler pole functions
    
    Should remains in rad/yr
    
    Written by C. Geisert (ENSTA/LIENSs) - 2017
    """
    if (incvn_ITRF is None) or (incve_ITRF is None):
        incvn_ITRF = np.ones(len(vn_ITRF))
        incve_ITRF = np.ones(len(ve_ITRF))
        
    all_pos = np.column_stack((lat_ITRF,long_ITRF))
    nstation=len(all_pos)
    vel_rot=[0]*nstation
    for i in range(nstation): #
        mat=Rt*np.array([np.array([np.sin(all_pos[i][1]*np.pi/180), -np.cos(all_pos[i][1]*np.pi/180),0]),
                         np.array([-np.sin(all_pos[i][0]*np.pi/180)*np.cos(all_pos[i][1]*np.pi/180),
                        -np.sin(all_pos[i][0]*np.pi/180)*np.sin(all_pos[i][1]*np.pi/180),np.cos(all_pos[i][0]*np.pi/180)])]) # [°] > [rad] pour le calcul
        
        
        vel_rot[i]=mat.dot(w)  #   [m] cross [rad/yr]
    #vel_rot=np.array(vel_rot)*1000  # [m/an]  > [mm/an]
    v_ITRF=np.transpose(np.array([vn_ITRF,ve_ITRF])) # les vitesses de toutes les stations [m/yr]
    vel_reltoref= v_ITRF-vel_rot  # [m/an]
    
    return vel_reltoref

def euler_pole_vector_from_latlongrate(wlat,wlong,wrate,
                                return_w_in_deg_per_Myr=False):
    """
    Compute the Euler vector from the pole latitude, longitude and rate
    
    Parameters
    ----------
    wlat,wlong : float
        latitude and longitude of the Euler pole (deg)

    wrate : float
        rate of the Euler pole (rad/Myr)
        
    Returns
    -------
    w : numpy.array
        Pole of rotation vector (rad/yr)

    Notes
    -----
    w is the common element for all Euler pole functions
    
    Should remains in rad/yr per default
    
    Written by C. Geisert (ENSTA/LIENSs) - 2017
    """
    # OUTPUT : w [rad/year]
     
    ### Conversion deg => rad coordinates
    wlong = np.deg2rad(wlong)    
    wlat  = np.deg2rad(wlat)
    
    wx=(wrate/1e6)*np.cos(wlat)*np.cos(wlong)#*np.pi/180  # [rad/yr]     
    wy=(wrate/1e6)*np.cos(wlat)*np.sin(wlong)#*np.pi/180  # [rad/yr]              
    wz=(wrate/1e6)*np.sin(wlat)#*np.pi/180                # [rad/yr]      
    w=[0]*3
    
    if return_w_in_deg_per_Myr:
        k = (180/np.pi) * 10**6
    else:
        k = 1.
    
    w[0]=wx*k
    w[1]=wy*k 
    w[2]=wz*k 
    
    return w

#%% ____Estimation de l'incertitude lors du calcul des EPP
def euler_pole_quality(w,vn_ref,ve_ref,nrmatinv,desmat,wwmat,
                       pretty_output=True):
    """ 
    Compute the uncertainties of the Euler pole determination

    Parameters
    ----------
    w : numpy.array
        Pole of rotation vector computed by euler_pole_calc 

    vn_ref,ve_ref : list or numpy.array
        north and east velocities of the reference points (m/yr)
        
    nrmatinv : 2-tuple
        output of scipy's lstsq fct from euler_pole_calc 
        
    wwmat : numpy.array
        weight matrix from euler_pole_calc 

    desmat : numpy.array
        design matrix from euler_pole_calc 
        
    pretty_output : bool
        if True, convert sigma_ww_latlon to pertinent units directly,
        returns raw units instead
                
    Returns
    -------
    sigma_ww : numpy.array
        Uncertainty on the Euler vector

    sigma_ww_latlon : numpy.array
        Uncertainty on the Euler pole : [rateSigma, latSigma, longSigma] 
        if pretty_output == True :  [deg/Myr,deg,deg] 
        if pretty_output == False : [rad/yr,rad,rad]
                
    dV_topo3 : numpy.array
        Residual velocities for the references points
        
    wrmse : float
        weigthed RMS on Residual velocities (m)
        
    wrmse_norm : float
        nomalized weigthed RMS on Residual velocities (m)
        
    rmse : float
        unweigthed RMS on Residual velocities (m)
        
    apost_sigma : float
        a-posteriori sigma (m)
        
    Notes
    -----
    w is the common element for all Euler pole functions
    
    Should remains in rad/yr
    """
    v_ref=np.transpose(np.array([vn_ref,ve_ref])) # les vitesses de toutes les stations [mm/yr]
    estm_vel=desmat.dot(w)
    estm_vel=np.transpose(np.array( [estm_vel[0::2],estm_vel[1::2]]))
    estm_vel_diff = v_ref- (estm_vel)  
    dV_topo= np.transpose(estm_vel_diff) 
    dV_topo=np.transpose(dV_topo)
    nd=len(dV_topo) 
    dV_topo=np.append(dV_topo[:,0],dV_topo[:,1])
    dV_topo2=(np.zeros([2*nd])) 
    for  index in range (0,2*nd-1,2):   
            dV_topo2[index]   = dV_topo[int(index/2)]
            dV_topo2[index+1] = dV_topo[int(index/2+nd)]                  
    df_value = len(dV_topo2) - 3;
    s0_2     = (np.transpose(dV_topo2).dot(wwmat).dot(dV_topo2)) / df_value;
    c_ww     = s0_2*(nrmatinv[0])  
    omega_length = np.sqrt(np.transpose(w) .dot( w))  
    #H  
    H00=  w[0]/omega_length
    H01=  w[1]/omega_length
    H02=  w[2]/omega_length
    H10= (-w[0] * w[2]) / (omega_length**2 * np.sqrt(w[0]**2 + w[1]**2))  
    H11= (-w[1] * w[2]) / (omega_length**2 * np.sqrt(w[0]**2 + w[1]**2)) 
    H12= -np.sqrt(w[0]**2 + w[1]**2) / omega_length**2  
    H20= -w[1]/(w[0]**2 + w[1]**2)
    H21=  w[0]/(w[0]**2 + w[1]**2)
    H22= 0
    H=np.array([[H00,H01,H02],
                [H10,H11,H12],
                [H20,H21,H22]])     
    c_ww_latlong = H.dot( c_ww ).dot( np.transpose(H) )  
    
    ### OUTPUT GENERATION
    sigma_ww         = np.sqrt(np.diag(c_ww)) 
    sigma_ww_latlong = np.sqrt(np.diag(c_ww_latlong)) 
    if pretty_output:
        rateSigma=((sigma_ww_latlong[0])*1e6*180/np.pi) # [rad/yr]> [°/Myr]
        latSigma=((sigma_ww_latlong[1])*180/np.pi) # [rad] >[°]
        longSigma=((sigma_ww_latlong[2])*180/np.pi) # [rad]>[°]
        sigma_ww_latlong = np.array([rateSigma,latSigma,longSigma])
    dV_topo3 = np.reshape(dV_topo2,(int(len(dV_topo2)/2),2))
    
    wrmse      = np.sqrt((np.transpose(dV_topo2).dot(wwmat)).dot(dV_topo2) / len(dV_topo2))
    wrmse_norm = np.sqrt((np.transpose(dV_topo2).dot(wwmat/(np.sum(np.diag(wwmat))))) .dot(dV_topo2) / len(dV_topo2))
    rmse       = np.sqrt(np.transpose(dV_topo2).dot(dV_topo2) / len(dV_topo2))
    apost_sigma = np.sqrt(s0_2)

    return sigma_ww,sigma_ww_latlong,dV_topo3,wrmse,wrmse_norm,rmse,apost_sigma

def topo2dm(coords):
    """
    Internal fuction for euler_pole_calc
    """
    n  = len(coords);
    dm = np.zeros([2 * n,3]) 
    for  index in range (0,2*n-1,2):       
        dm[index,0]   =  np.sin(coords[int((index+1)/2)][1]);
        dm[index,1]   = -np.cos(coords[int((index+1)/2)][1]);  
        dm[index+1,0] = -np.sin(coords[int((index+1)/2)][0]) * np.cos(coords[int((index+1)/2)][1]);
        dm[index+1,1] = -np.sin(coords[int((index+1)/2)][0]) * np.sin(coords[int((index+1)/2)][1]);
        dm[index+1,2] =  np.cos(coords[int((index+1)/2)][0]);   
    return dm


def covarvec2wtmat(covars):
    """
    Internal fuction for euler_pole_calc
    """
    wtmat = []
    aa=np.shape(covars)
    cc = np.zeros([aa[0] * 2,aa[0] * 2]);
    a=np.shape(cc)
    for idx in range( 0 ,a[0],2):    
        cc[idx,idx]         = covars[int(idx/2)][0]**2;
        cc[idx + 1,idx + 1] = covars[int(idx/2)][1]**2;

        cc[idx,idx + 1]     = covars[int(idx/2)][2];
        cc[idx + 1,idx]     = covars[int(idx/2)][2];
    
    cc[np.isnan(cc)] = 0;  
    wtmat = np.linalg.lstsq(cc,np.eye(a[0]),rcond=None) 
    return wtmat  






